import os

from tito.builder import UpstreamBuilder
from tito.common import debug, run_command, error_out
from tito.compat import getstatusoutput


class DistributionBuilder(UpstreamBuilder):
    """ This class is used for building packages for distributions.
    Parent class UpstreamBuilder build one big patch from upstream and create e.g.:
      Patch0: foo-1.2.13-1-to-foo-1.2.13-3-sat.patch
    This class create one patch per each release. E.g.:
      Patch0: foo-1.2.13-1-to-foo-1.2.13-2-sat.patch
      Patch1: foo-1.2.13-2-to-foo-1.2.13-3-sat.patch
    """
    def __init__(self, name=None, tag=None, build_dir=None,
            config=None, user_config=None, args=None, **kwargs):
        UpstreamBuilder.__init__(self, name, tag, build_dir, config,
                user_config, args, **kwargs)
        self.patch_files = []

    def patch_upstream(self):
        """ Create one patch per each release """
        ch_dir = self.git_root
        if self.relative_project_dir != "/":
            ch_dir = os.path.join(self.git_root,
                    self.relative_project_dir)
        os.chdir(ch_dir)
        debug("Running generate-patches.pl -d %s %s %s-1 %s %s"
               % (self.rpmbuild_gitcopy, self.project_name, self.upstream_version, self.build_version, self.git_commit_id))
        output = run_command("generate-patches.pl -d %s %s %s-1 %s %s"
               % (self.rpmbuild_gitcopy, self.project_name, self.upstream_version, self.build_version, self.git_commit_id))
        self.patch_files = output.split("\n")
        for p_file in self.patch_files:
            (status, output) = getstatusoutput(
                "grep 'Binary files .* differ' %s/%s " % (self.rpmbuild_gitcopy, p_file))
            if status == 0 and output != "":
                print(output)
                error_out("You are doomed. Diff contains binary files. You can not use this builder")

            run_command("cp %s/%s %s" % (self.rpmbuild_gitcopy, p_file, self.rpmbuild_sourcedir))

        (patch_number, patch_insert_index, patch_apply_index, lines) = self._patch_upstream()

        for patch in self.patch_files:
            lines.insert(patch_insert_index, "Patch%s: %s\n" % (patch_number, patch))
            lines.insert(patch_apply_index, "%%patch%s -p1\n" % (patch_number))
            patch_number += 1
            patch_insert_index += 1
            patch_apply_index += 2
        self._write_spec(lines)



# Copyright (c) 2008-2009 Red Hat, Inc.
#
# This software is licensed to you under the GNU General Public License,
# version 2 (GPLv2). There is NO WARRANTY for this software, express or
# implied, including the implied warranties of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. You should have received a copy of GPLv2
# along with this software; if not, see
# http://www.gnu.org/licenses/old-licenses/gpl-2.0.txt.
#
# Red Hat trademarks are not licensed under GPLv2. No permission is
# granted to use or replicate Red Hat trademarks that are incorporated
# in this software or its documentation.

import re
import struct
import sys
import codecs
import tito.compat

RECORD_SIZE = 512

# Git writes its tarballs to be a multiple of 10240.  I'm not sure why: the
# implementation in archive-tar.c doesn't have any comments on the matter.
GIT_BLOCK_SIZE = RECORD_SIZE * 20


class TarFixer(object):
    """Code for updating a tar header's mtime.  For details on the tar format
    see http://www.gnu.org/software/tar/manual/html_node/Standard.html and
    http://en.wikipedia.org/wiki/Tar_%28computing%29

    Tito passes "git archive" a tree ID.  The "git archive" man page states:

        git archive behaves differently when given a tree ID versus when given
        a commit ID or tag ID. In the first case the current time is used as
        the modification time of each file in the archive.

    Using the current time means that every time we build the source tarball,
    the file fingerprint will change since the metadata in the tarball changes.
    We don't want that since build systems track the fingerprint to see if
    the actual source has changed.

    The resultant tarball will be in this format:

        - Global header (512 bytes)
        - Extended header block with git ref (512 bytes)
        - [File header (512 bytes) + File data padded to multiple of 512] * number of files
        - 1024 NUL bytes
        - However many NUL bytes are necessary to pad the file to a multiple of GIT_BLOCK_SIZE

    The block after the global header with the git ref is called an "extended header".
    We are technically writing a "pax" archive because of the use of extensions.  According
    to the comments in git's archive-tar.c

        pax extended header records have the format "%u %s=%s\n".  %u contains
        the size of the whole string (including the %u), the first %s is the
        keyword, the second one is the value.

    PAX (also known as POSIX.1-2001) always encodes everything in UTF-8.
    """
    def __init__(self, fh, out, timestamp, gitref, maven_built=False):
        self.maven_built = maven_built

        # As defined in tar.h
        # An collections.OrderedDict would be more appropriate here but I'm trying to
        # maintain Python 2.6 compatibility.
        self.tar_struct = [
            ('name', '100s'),
            ('mode', '8s'),
            ('uid', '8s'),
            ('gid', '8s'),
            ('size', '12s'),
            ('mtime', '12s'),
            ('checksum', '8s'),
            ('typeflag', '1s'),
            ('linkname', '100s'),
            ('magic', '6s'),
            ('version', '2s'),
            ('uname', '32s'),
            ('gname', '32s'),
            ('devmajor', '8s'),
            ('devminor', '8s'),
            ('prefix', '155s'),
        ]

        # The items in the list below are zero-padded octal numbers in ASCII.
        # All other fields are null-terminated character strings. Each numeric
        # field of width w contains w minus 1 digits, and a null.
        #
        # The checksum is technically an octal_member but we handle it specially.
        self.octal_members = [
            'mode',
            'uid',
            'gid',
            'size',
            'mtime',
            'devmajor',
            'devminor',
        ]

        # Add an '=' to use native byte order with standard sizes
        self.struct_template = "=" + "".join(map(lambda x: x[1], self.tar_struct))
        self.struct_members = list(map(lambda x: x[0], self.tar_struct))
        self.struct_hash = dict(self.tar_struct)

        # The tarballs created by git archive from tree IDs don't have a global
        # header for some reason.
        self.need_header = True
        self.done = False

        # We need to track the total number of bytes we've written so we can
        # pad out the final tarball to be a multiple of GIT_BLOCK_SIZE
        self.total_length = 0

        self.fh = fh
        self.out = out
        self.timestamp = int(timestamp)
        self.gitref = gitref

    def full_read(self, read_size):
        read = self.fh.read(read_size)
        amount_read = len(read)
        while amount_read < read_size:
            left_to_read = read_size - amount_read
            next_read = self.fh.read(left_to_read)

            if next_read == '':
                raise IOError("Buffer underflow when reading")

            amount_read += len(next_read)
            read = read + next_read

        return read

    def write(self, data):
        self.out.write(tito.compat.ensure_binary(data))

    def chunk_to_hash(self, chunk):
        # Our struct template is only 500 bytes, but the last 12 bytes are NUL
        # I elected to ignore them completely instead of including them in the
        # template as '12x'.  The unpack_from method will read the bytes our
        # template defines from chunk and discard the rest.
        unpacked = struct.unpack_from(self.struct_template, chunk)
        unpacked = list(map(lambda x: tito.compat.ensure_text(x), unpacked))
        # Zip what we read together with the member names and create a dictionary
        chunk_props = dict(zip(self.struct_members, unpacked))

        return chunk_props

    def padded_size(self, length, pad_size=RECORD_SIZE):
        """Function to pad out a length to the nearest multiple of pad_size
        that can contain it."""
        blocks = length // pad_size
        if length % pad_size != 0:
            blocks += 1
        return blocks * pad_size

    def create_global_header(self):
        header_props = {
            'name': u'pax_global_header',
            'mode': 0o666,
            'uid': 0,
            'gid': 0,
            'size': 52,  # The size of the extended header with the gitref
            'mtime': self.timestamp,
            'typeflag': u'g',
            'linkname': u'',
            'magic': u'ustar',
            'version': u'00',
            'uname': u'root',
            'gname': u'root',
            'devmajor': 0,
            'devminor': 0,
            'prefix': u'',
        }
        self.process_header(header_props)

    def encode_header(self, chunk_props, encode_order=None):
        pack_values = []
        if encode_order is None:
            encode_order = self.struct_members

        for member in encode_order:
            if member in self.octal_members:
                # Pad out the octal value to the right length
                member_template = self.struct_hash[member]
                field_size = int(re.match('(\d+)', member_template).group(1)) - 1
                fmt = "%0" + str(field_size) + "o\x00"
                as_string = fmt % chunk_props[member]
                pack_values.append(tito.compat.ensure_binary(as_string))
            else:
                pack_values.append(tito.compat.ensure_binary(chunk_props[member]))
        return pack_values

    def process_header(self, chunk_props):
        """There is a header before every file and a global header at the top."""
        chunk_props['checksum'] = self.calculate_checksum(chunk_props)
        pack_values = self.encode_header(chunk_props)

        # The struct itself is only 500 bytes so we have to pad it to 512
        data_out = struct.pack(self.struct_template + "12x", *pack_values)
        self.write(data_out)
        self.total_length += len(data_out)

    def process_extended_header(self):
        # Trash the original comment
        self.full_read(RECORD_SIZE)
        self.create_extended_header()

    def create_extended_header(self):
        # pax extended header records have the format "%u %s=%s\n".  %u contains
        # the size of the whole string (including the %u), the first %s is the
        # keyword, the second one is the value.
        #
        # Since the git ref is always 40 ASCII characters we can pre-compute the length
        # to put in the extended header
        comment = "52 comment=%s\n" % self.gitref
        data_out = struct.pack("=52s460x", tito.compat.ensure_binary(comment, "ascii"))
        self.write(data_out)
        self.total_length += len(data_out)

    def process_file_data(self, size):
        data_out = self.full_read(self.padded_size(size))
        self.write(data_out)
        self.total_length += len(data_out)

    def calculate_checksum(self, chunk_props):
        """The checksum field is the ASCII representation of the octal value of the simple
        sum of all bytes in the header block. Each 8-bit byte in the header is added
        to an unsigned integer, initialized to zero, the precision of which shall be
        no less than seventeen bits. When calculating the checksum, the checksum field is
        treated as if it were all spaces.
        """
        chunk_props['checksum'] = " " * 8
        values = self.encode_header(chunk_props)
        new_chksum = 0
        for val in values:
            val_bytes = bytearray(tito.compat.ensure_binary(val))
            for b in val_bytes:
                new_chksum += b
        return "%07o\x00" % new_chksum

    def process_chunk(self, chunk):
        # Tar archives end with two 512 byte blocks of zeroes
        if chunk == b"\x00" * 512:
            self.write(b"\x00" * 512)
            self.total_length += len(chunk)
            if self.last_chunk_was_nulls:
                final_padding = b"\x00" * (self.padded_size(self.total_length, GIT_BLOCK_SIZE) - self.total_length)
                self.write(final_padding)
                self.done = True
            self.last_chunk_was_nulls = True
            return

        self.last_chunk_was_nulls = False

        chunk_props = self.chunk_to_hash(chunk)

        # Delete the old checksum since it's now invalid and we don't want even
        # an inadvertent reference to it.
        del(chunk_props['checksum'])

        # Remove the trailing NUL byte(s) on the end of members
        for k, v in chunk_props.items():
            chunk_props[k] = v.rstrip("\x00")

        # This line is the whole purpose of this class!
        chunk_props['mtime'] = "%o" % self.timestamp

        if self.maven_built:
            # Maven does all sorts of horrible things in the tarfile it creates.
            # Everything is padded out with spaces instead of NUL bytes and the uid
            # and gid fields are left empty.
            #
            # Plus it sets the uname and gname to the current user resulting in
            # the checksum changing from person to person.
            # See https://jira.codehaus.org/browse/PLXCOMP-233
            chunk_props['uname'] = 'root'
            chunk_props['gname'] = 'root'
            chunk_props['uid'] = '0'
            chunk_props['gid'] = '0'
            # In a tar file, the highest 3 bits in the mode represent if the tarfile
            # should be extracted with the GID or UID set.  Maven adds these but we don't
            # want them, so we just take the last 4 which are the ones that matter to us.
            chunk_props['mode'] = str(chunk_props['mode'])[-4:-1]
            chunk_props['version'] = '00'
            for x in ['size', 'devmajor', 'devminor']:
                chunk_props[x] = chunk_props[x].strip()

        for member in self.octal_members:
            # Convert octals to decimal
            chunk_props[member] = int(chunk_props[member], 8)

        # If there is no global header, we need to create one
        if self.need_header:
            # When run against a tree ID, git archive doesn't create
            # a global header.  The first block is just the header for
            # the first file.
            if chunk_props['typeflag'] != 'g':
                self.create_global_header()
                self.create_extended_header()
                self.process_header(chunk_props)
            else:
                self.process_header(chunk_props)
                self.process_extended_header()
            self.need_header = False
        else:
            self.process_header(chunk_props)
            self.process_file_data(chunk_props['size'])

    def fix(self):
        # The gzip file object has its mode as an integer.  We have to
        # access the underlying file object to get the real mode.
        if hasattr(self.fh, "myfileobj"):
            mode = self.fh.myfileobj.mode
        else:
            mode = self.fh.mode

        if 'b' not in mode:
            raise IOError("The input file must be opened in binary mode!")

        try:
            chunk = self.full_read(RECORD_SIZE)
            while chunk != "" and not self.done:
                self.process_chunk(chunk)
                if not self.done:
                    chunk = self.full_read(RECORD_SIZE)
        finally:
            self.fh.close()


if __name__ == '__main__':
    if len(sys.argv) != 5:
        sys.exit("Usage: %s UNIX_TIMESTAMP GIT_HASH TAR_FILE DESTINATION_FILE" % sys.argv[0])

    try:
        timestamp = int(sys.argv[1])
    except:
        sys.exit("UNIX_TIMESTAMP must be an integer")

    gitref = sys.argv[2]
    tar_file = sys.argv[3]
    destination_file = sys.argv[4]

    try:
        dfh = open(destination_file, 'wb')
    except:
        print("Could not open %s" % destination_file)

    try:
        fh = open(tar_file, 'rb')
    except:
        print("Could not read %s" % tar_file)

    reader = TarFixer(fh, dfh, timestamp, gitref)
    reader.fix()



# Copyright (c) 2008-2011 Red Hat, Inc.
#
# This software is licensed to you under the GNU General Public License,
# version 2 (GPLv2). There is NO WARRANTY for this software, express or
# implied, including the implied warranties of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. You should have received a copy of GPLv2
# along with this software; if not, see
# http://www.gnu.org/licenses/old-licenses/gpl-2.0.txt.
#
# Red Hat trademarks are not licensed under GPLv2. No permission is
# granted to use or replicate Red Hat trademarks that are incorporated
# in this software or its documentation.

import os
import tempfile
import subprocess
import sys

from tito.common import run_command, debug
from tito.compat import getoutput, write, getstatusoutput
from tito.release.distgit import FedoraGitReleaser
from tito.bugtracker import BugzillaExtractor


class ObsReleaser(FedoraGitReleaser):
    """ Releaser for Open Build System using osc command """

    REQUIRED_CONFIG = ['project_name']
    cli_tool = "osc"

    def __init__(self, name=None, tag=None, build_dir=None,
            config=None, user_config=None,
            target=None, releaser_config=None, no_cleanup=False,
            test=False, auto_accept=False, **kwargs):
        FedoraGitReleaser.__init__(self, name, tag, build_dir,
                user_config, target, releaser_config, no_cleanup, test, auto_accept)

        self.obs_project_name = \
            self.releaser_config.get(self.target, "project_name")

        if self.config.has_option(self.target, "package_name"):
            self.obs_package_name = self.config.get(self.target, "package_name")
        else:
            self.obs_package_name = self.project_name

        self.package_workdir = os.path.join(self.working_dir, self.obs_project_name,
                self.project_name)

    def release(self, dry_run=False, no_build=False, scratch=False):
        self.dry_run = dry_run
        self.no_build = no_build

        getoutput("mkdir -p %s" % self.working_dir)
        os.chdir(self.working_dir)
        run_command("%s co %s %s" % (self.cli_tool, self.obs_project_name, self.obs_package_name))

        os.chdir(self.package_workdir)

        self.builder.tgz()
        if self.test:
            self.builder._setup_test_specfile()

        self._obs_sync_files(self.package_workdir)
        self._obs_user_confirm_commit(self.package_workdir)

    def _confirm_commit_msg(self, diff_output):
        """
        Generates a commit message in a temporary file, gives the user a
        chance to edit it, and returns the filename to the caller.
        """

        fd, name = tempfile.mkstemp()
        debug("Storing commit message in temp file: %s" % name)
        write(fd, "Update %s to %s\n" % (self.obs_package_name,
            self.builder.build_version))
        # Write out Resolves line for all bugzillas we see in commit diff:
        extractor = BugzillaExtractor(diff_output)
        for line in extractor.extract():
            write(fd, line + "\n")

        print("")
        print("##### Commit message: #####")
        print("")

        os.lseek(fd, 0, 0)
        commit_file = os.fdopen(fd)
        for line in commit_file.readlines():
            print(line)
        commit_file.close()

        print("")
        print("###############################")
        print("")
        if self._ask_yes_no("Would you like to edit this commit message? [y/n] ", False):
            debug("Opening editor for user to edit commit message in: %s" % name)
            editor = 'vi'
            if "EDITOR" in os.environ:
                editor = os.environ["EDITOR"]
            subprocess.call(editor.split() + [name])

        return name

    def _obs_user_confirm_commit(self, project_checkout):
        """ Prompt user if they wish to proceed with commit. """
        print("")
        text = "Running '%s diff' in: %s" % (self.cli_tool, project_checkout)
        print("#" * len(text))
        print(text)
        print("#" * len(text))
        print("")

        os.chdir(project_checkout)

        (status, diff_output) = getstatusoutput("%s diff" % self.cli_tool)

        if diff_output.strip() == "":
            print("No changes in main branch, skipping commit.")
        else:
            print(diff_output)
            print("")
            print("##### Please review the above diff #####")
            if not self._ask_yes_no("Do you wish to proceed with commit? [y/n] "):
                print("Fine, you're on your own!")
                self.cleanup()
                sys.exit(1)

            print("Proceeding with commit.")
            commit_msg_file = self._confirm_commit_msg(diff_output)
            cmd = '%s commit -F %s' % (self.cli_tool,
                    commit_msg_file)
            debug("obs commit command: %s" % cmd)
            print
            if self.dry_run:
                self.print_dry_run_warning(cmd)
            else:
                print("Proceeding with commit.")
                os.chdir(self.package_workdir)
                print(run_command(cmd))

            os.unlink(commit_msg_file)

        if self.no_build:
            getstatusoutput("%s abortbuild %s %s" % (
                self.cli_tool, self.obs_project_name, self.obs_package_name))
            print("Aborting automatic rebuild because --no-build has been specified.")

    def _obs_sync_files(self, project_checkout):
        """
        Copy files from our obs checkout into each obs checkout and add them.

        A list of safe files is used to protect critical files both from
        being overwritten by a osc file of the same name, as well as being
        deleted after.
        """

        # Build the list of all files we will copy:
        debug("Searching for files to copy to build system osc checkout:")
        files_to_copy = self._list_files_to_copy()

        os.chdir(project_checkout)

        self._sync_files(files_to_copy, project_checkout)

        os.chdir(project_checkout)

        # Add/remove everything:
        run_command("%s addremove" % (self.cli_tool))



# flake8: noqa

from tito.release.main import \
    Releaser, \
    RsyncReleaser, \
    YumRepoReleaser, \
    KojiReleaser, \
    KojiGitReleaser

from tito.release.distgit import FedoraGitReleaser, DistGitReleaser, DistGitMeadReleaser, CentosGitReleaser
from tito.release.obs import ObsReleaser
from tito.release.copr import CoprReleaser



import os
import re
from bugzilla.rhbugzilla import RHBugzilla
from tito.common import debug, error_out
from tito.exception import TitoException
from tito.compat import xmlrpclib


class MissingBugzillaCredsException(TitoException):
    pass


class BugzillaExtractor(object):
    """
    Parses output of a dist-git commit diff looking for changelog
    entries that look like they reference bugzilla commits.

    Optionally can check bugzilla for required flags on each bug.
    """
    def __init__(self, diff_output, required_flags=None,
        placeholder_bz=None):

        self.diff_output = diff_output
        self.required_flags = required_flags
        self.placeholder_bz = placeholder_bz

        # Tuples of bugzilla ID + commit message we extracted:
        self.bzs = []

    def extract(self):

        self.bzs = self._extract_bzs()

        if self.required_flags:
            self._check_for_bugzilla_creds()
            self.bzs = self._filter_bzs_with_flags()

        return self._format_lines()

    def _check_for_bugzilla_creds(self):
        if not os.path.exists(os.path.expanduser("~/.bugzillarc")):
            raise MissingBugzillaCredsException("Missing ~/.bugzillarc")
        else:
            debug("Found bugzilla credentials in ~/.bugzillarc")

    def _extract_bzs(self):
        """
        Parses the output of CVS diff or a series of git commit log entries,
        looking for new lines which look like a commit of the format:

        ######: Commit message

        Returns a list of lines of text similar to:

        Resolves: #XXXXXX - Commit message

        If the releaser specifies any required bugzilla flags we will
        check each bug found and see if it has all required flags. If not
        we skip it. If we end up with *no* bugs with the required flags
        our build is likely to fail, so we look for a placeholder bugzilla
        defined in relaser config and use that instead if possible, otherwise
        error out.

        Returns a list of lines to write to the commit message as is.
        """
        regex = re.compile(r"^- (\d*)\s?[:-]+\s?(.*)")
        diff_regex = re.compile(r"^(\+- )+(\d*)\s?[:-]+\s?(.*)")
        bzs = []
        for line in self.diff_output.split("\n"):
            match = re.match(regex, line)
            match2 = re.match(diff_regex, line)
            if match:
                bzs.append((match.group(1), match.group(2)))
            elif match2:
                bzs.append((match2.group(2), match2.group(3)))
        return bzs

    def _format_lines(self):
        output = []
        for bz in self.bzs:
            output.append("Resolves: #%s - %s" % (bz[0], bz[1]))
        if len(output) == 0 and self.required_flags:
            # No bugzillas had required flags, use a placeholder if
            # we have one, otherwise we have to error out.
            if self.placeholder_bz:
                print("No bugs with required flags were found, using placeholder: %s" % self.placeholder_bz)
                output.append("Related: #%s" % self.placeholder_bz)
            else:
                error_out("No bugzillas found with required flags: %s" %
                    self.required_flags)
        return output

    def _filter_bzs_with_flags(self):
        print("Checking flags on bugs: %s" % self.bzs)
        print("  required flags: %s" % self.required_flags)

        # TODO: Would be nice to load bugs in bulk here but for now we'll
        # keep it simple.
        filtered_bzs = []
        for bz_tuple in self.bzs:
            bug_id = bz_tuple[0]
            try:
                bug = self._load_bug(bug_id)
            except xmlrpclib.Fault:
                print("WARNING: Bug %s does not seem to exist." % bug_id)
                continue
            debug("Bug %s has flags: %s" % (bug_id, bug.flags))
            flags_missing = False
            for flag in self.required_flags:
                if bug.get_flag_status(flag[0:-1]) != flag[-1]:
                    print("WARNING: Bug %s missing required flag: %s" %
                        (bug_id, flag))
                    flags_missing = True
                    break
                else:
                    debug("Bug %s has required flag: %s" %
                        (bug_id, flag))
            if not flags_missing:
                filtered_bzs.append(bz_tuple)
        return filtered_bzs

    def _load_bug(self, bug_id):
        bugzilla = RHBugzilla(url='https://bugzilla.redhat.com/xmlrpc.cgi')
        return bugzilla.getbug(bug_id, include_fields=['id', 'flags'])



# Copyright (c) 2008-2009 Red Hat, Inc.
#
# This software is licensed to you under the GNU General Public License,
# version 2 (GPLv2). There is NO WARRANTY for this software, express or
# implied, including the implied warranties of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. You should have received a copy of GPLv2
# along with this software; if not, see
# http://www.gnu.org/licenses/old-licenses/gpl-2.0.txt.
#
# Red Hat trademarks are not licensed under GPLv2. No permission is
# granted to use or replicate Red Hat trademarks that are incorporated
# in this software or its documentation.
"""
Code for tagging Spacewalk/Satellite packages.
"""

import os
import re
import rpm
import shutil
import subprocess
import sys
import tempfile
import textwrap

try:
    from shlex import quote
except ImportError:
    from pipes import quote

from string import Template

from time import strftime

from tito.common import (debug, error_out, run_command,
        find_spec_like_file, get_project_name, get_latest_tagged_version,
        get_spec_version_and_release, replace_version,
        tag_exists_locally, tag_exists_remotely, head_points_to_tag, undo_tag,
        increase_version, reset_release, increase_zstream, warn_out,
        BUILDCONFIG_SECTION, get_relative_project_dir_cwd, info_out,
        get_git_user_info)
from tito.compat import write, StringIO, getstatusoutput
from tito.exception import TitoException
from tito.config_object import ConfigObject
from tito.tagger.cargobump import CargoBump


class VersionTagger(ConfigObject):
    """
    Standard Tagger class, used for tagging packages built from source in
    git. (as opposed to packages which commit a tarball directly into git).

    Releases will be tagged by incrementing the package version,
    and the actual RPM "release" will always be set to 1.
    """

    def __init__(self, config=None, keep_version=False, offline=False, user_config=None):
        ConfigObject.__init__(self, config=config)
        self.user_config = user_config

        self.full_project_dir = os.getcwd()
        self.spec_file_name = find_spec_like_file()
        self.project_name = get_project_name(tag=None)

        self.relative_project_dir = get_relative_project_dir_cwd(
            self.git_root)  # i.e. java/

        self.spec_file = os.path.join(self.full_project_dir,
                self.spec_file_name)
        self.keep_version = keep_version

        self.today = self._changelog_date()
        (self.git_user, self.git_email) = get_git_user_info()
        git_email = self.git_email
        if git_email is None:
            git_email = ''
        self.changelog_regex = re.compile('\\*\s%s\s%s(\s<%s>)?' % (self.today,
            self.git_user, git_email.replace("+", "\+").replace(".", "\.")))

        self._no_auto_changelog = False
        self._accept_auto_changelog = False
        self._new_changelog_msg = "new package built with tito"
        self._changelog = None
        self.offline = offline

    def run(self, options):
        """
        Perform the actions requested of the tagger.

        NOTE: this method may do nothing if the user requested no build actions
        be performed. (i.e. only release tagging, etc)
        """
        if options.tag_release:
            warn_out("--tag-release option no longer necessary,"
                " 'tito tag' will accomplish the same thing.")
        if options.no_auto_changelog:
            self._no_auto_changelog = True
        if options.accept_auto_changelog:
            self._accept_auto_changelog = True
        if options.auto_changelog_msg:
            self._new_changelog_msg = options.auto_changelog_msg
        if options.use_version:
            self._use_version = options.use_version
        if options.use_release:
            self._use_release = options.use_release
        if options.changelog:
            self._changelog = options.changelog

        self.check_tag_precondition()

        # Only two paths through the tagger module right now:
        if options.undo:
            self._undo()
        else:
            self._tag_release()

    def check_tag_precondition(self):
        if self.config.has_option("tagconfig", "require_package"):
            packages = self.config.get("tagconfig", "require_package").split(',')
            ts = rpm.TransactionSet()
            missing_packages = []
            for p in packages:
                p = p.strip()
                mi = ts.dbMatch('name', p)
                if not mi:
                    missing_packages.append(p)
            if missing_packages:
                raise TitoException("To tag this package, you must first install: %s" %
                    ', '.join(missing_packages))

    def _tag_release(self):
        """
        Tag a new version of the package. (i.e. x.y.z+1)
        """
        self._make_changelog()
        new_version = self._bump_version()
        self._check_tag_does_not_exist(self._get_new_tag(new_version))
        self._update_changelog(new_version)
        CargoBump.tag_new_version(self.full_project_dir, new_version)
        self._update_setup_py(new_version)
        self._update_pom_xml(new_version)
        self._update_package_metadata(new_version)

    def _undo(self):
        """
        Undo the most recent tag.

        Tag commit must be the most recent commit, and the tag must not
        exist in the remote git repo, otherwise we report and error out.
        """
        tag = self._get_tag_for_version(get_latest_tagged_version(self.project_name))
        info_out("Undoing tag: %s" % tag)
        if not tag_exists_locally(tag):
            raise TitoException(
                "Cannot undo tag that does not exist locally.")
        if not self.offline and tag_exists_remotely(tag):
            raise TitoException("Cannot undo tag that has been pushed.")

        # Tag must be the most recent commit.
        if not head_points_to_tag(tag):
            raise TitoException("Cannot undo if tag is not the most recent commit.")

        # Everything looks good:
        print
        undo_tag(tag)

    def _changelog_date(self):
        option = (BUILDCONFIG_SECTION, "changelog_date_with_time")
        if self.config.has_option(*option) and self.config.getboolean(*option):
            return strftime("%a %b %d %T %Z %Y")
        return strftime("%a %b %d %Y")

    def _changelog_remove_cherrypick(self, line):
        """
        remove text "(cherry picked from commit ..." from line unless
        changelog_do_not_remove_cherrypick is specified in [BUILDCONFIG_SECTION]
        """
        if not (self.config.has_option(BUILDCONFIG_SECTION, "changelog_do_not_remove_cherrypick")
            and self.config.get(BUILDCONFIG_SECTION, "changelog_do_not_remove_cherrypick")
            and self.config.get(BUILDCONFIG_SECTION, "changelog_do_not_remove_cherrypick").strip() != '0'):
            m = re.match("(.+)(\(cherry picked from .*\))", line)
            if m:
                line = m.group(1)
        return line

    def _changelog_format(self):
        """
        If you have set changelog_format in [BUILDCONFIG_SECTION], it will return
        that string.  Otherwise, return one of two defaults:

        - '%s (%ae)', if changelog_with_email is unset or evaluates to True
        - '%s', if changelog_with_email is set and evaluates to False
        """
        result = ''
        if self.config.has_option(BUILDCONFIG_SECTION, "changelog_format"):
            result = self.config.get(BUILDCONFIG_SECTION, "changelog_format")
        else:
            with_email = ''
            if (self.config.has_option(BUILDCONFIG_SECTION, "changelog_with_email")
                and (self.config.get(BUILDCONFIG_SECTION, "changelog_with_email")) not in ['0', '']) or \
                not self.config.has_option(BUILDCONFIG_SECTION, "changelog_with_email"):
                with_email = ' (%ae)'
            result = "%%s%s" % with_email
        return result

    def _generate_default_changelog(self, last_tag):
        """
        Run git-log and will generate changelog, which still can be edited by user
        in _make_changelog.
        """
        patch_command = "git log"
        if self.config.has_option(BUILDCONFIG_SECTION, "keep_merge_commits"):
            keep = self.config.get(BUILDCONFIG_SECTION, "keep_merge_commits")
            if not keep or keep.strip().lower() not in ['1', 'true']:
                patch_command += " --no-merges"
        else:
            patch_command += " --no-merges"
        patch_command += " --pretty='format:%s' --relative %s..%s -- %s" % (
            self._changelog_format(), last_tag, "HEAD", ".")
        output = run_command(patch_command)
        result = []
        for line in output.split('\n'):
            line = line.replace('%', '%%')
            result.extend([self._changelog_remove_cherrypick(line)])
        return '\n'.join(result)

    def _make_changelog(self):
        """
        Create a new changelog entry in the spec, with line items from git
        """
        if self._no_auto_changelog:
            debug("Skipping changelog generation.")
            return

        in_f = open(self.spec_file, 'r')
        out_f = open(self.spec_file + ".new", 'w')

        found_changelog = False
        for line in in_f.readlines():
            out_f.write(line)

            if not found_changelog and line.startswith("%changelog"):
                found_changelog = True

                old_version = get_latest_tagged_version(self.project_name)

                fd, name = tempfile.mkstemp()
                write(fd, "# Create your changelog entry below:\n")
                if self.git_email is None or (('HIDE_EMAIL' in self.user_config) and
                        (self.user_config['HIDE_EMAIL'] not in ['0', ''])):
                    header = "* %s %s\n" % (self.today, self.git_user)
                else:
                    header = "* %s %s <%s>\n" % (self.today, self.git_user,
                       self.git_email)

                write(fd, header)

                # don't die if this is a new package with no history
                if self._changelog is not None:
                    for entry in self._changelog:
                        if not entry.startswith('-'):
                            entry = '- ' + entry
                        write(fd, entry)
                        write(fd, "\n")
                else:
                    if old_version is not None:
                        last_tag = self._get_new_tag(old_version)
                        output = self._generate_default_changelog(last_tag)
                    else:
                        output = self._new_changelog_msg

                    for cmd_out in output.split("\n"):
                        write(fd, "- ")
                        write(fd, "\n  ".join(textwrap.wrap(cmd_out, 77)))
                        write(fd, "\n")

                write(fd, "\n")

                if not self._accept_auto_changelog:
                    # Give the user a chance to edit the generated changelog:
                    editor = 'vi'
                    if "EDITOR" in os.environ:
                        editor = os.environ["EDITOR"]
                    subprocess.call(editor.split() + [name])

                os.lseek(fd, 0, 0)
                f = os.fdopen(fd)

                for line in f.readlines():
                    if not line.startswith("#"):
                        out_f.write(line)

                output = f.read()

                f.close()
                os.unlink(name)

        if not found_changelog:
            warn_out("no %changelog section find in spec file. Changelog entry was not appended.")

        in_f.close()
        out_f.close()

        shutil.move(self.spec_file + ".new", self.spec_file)

    def _update_changelog(self, new_version):
        """
        Update the changelog with the new version.
        """
        # Not thrilled about having to re-read the file here but we need to
        # check for the changelog entry before making any modifications, then
        # bump the version, then update the changelog.
        f = open(self.spec_file, 'r')
        buf = StringIO()
        found_match = False
        for line in f.readlines():
            match = self.changelog_regex.match(line)
            if match and not found_match:
                buf.write("%s %s\n" % (match.group(), new_version))
                found_match = True
            else:
                buf.write(line)
        f.close()

        # Write out the new file contents with our modified changelog entry:
        f = open(self.spec_file, 'w')
        f.write(buf.getvalue())
        f.close()
        buf.close()

    def _update_setup_py(self, new_version):
        """
        If this project has a setup.py, attempt to update it's version.
        """
        self._update_version_file(new_version)

        setup_file = os.path.join(self.full_project_dir, "setup.py")
        if not os.path.exists(setup_file):
            return

        debug("Found setup.py, attempting to update version.")

        # We probably don't want version-release in setup.py as release is
        # an rpm concept. Hopefully this assumption on
        py_new_version = new_version.split('-')[0]

        f = open(setup_file, 'r')
        buf = StringIO()
        for line in f.readlines():
            buf.write(replace_version(line, py_new_version))
        f.close()

        # Write out the new setup.py file contents:
        f = open(setup_file, 'w')
        f.write(buf.getvalue())
        f.close()
        buf.close()

        run_command("git add %s" % setup_file)

    def _update_pom_xml(self, new_version):
        """
        If this project uses Maven, attempt to update the version in pom.xml
        """
        # Remove the release since Maven doesn't understand that really

        pom_file = os.path.join(self.full_project_dir, "pom.xml")
        if not os.path.exists(pom_file):
            return

        mvn_new_version = new_version.split('-')[0]

        maven_args = ['-B']
        if 'MAVEN_ARGS' in self.user_config:
            maven_args.append(self.user_config['MAVEN_ARGS'])
        else:
            maven_args.append('-q')

        run_command("mvn %s versions:set -DnewVersion=%s -DgenerateBackupPoms=false" % (
            " ".join(maven_args),
            mvn_new_version))
        run_command("git add %s" % pom_file)

    def _bump_version(self, release=False, zstream=False):
        """
        Bump up the package version in the spec file.

        Set release to True to bump the package release instead.

        Checks for the keep version option and if found, won't actually
        bump the version or release.
        """
        old_version = get_latest_tagged_version(self.project_name)
        if old_version is None:
            old_version = "untagged"
        if not self.keep_version:
            version_regex = re.compile("^(version:\s*)(.+)$", re.IGNORECASE)
            release_regex = re.compile("^(release:\s*)(.+)$", re.IGNORECASE)

            in_f = open(self.spec_file, 'r')
            out_f = open(self.spec_file + ".new", 'w')

            for line in in_f.readlines():
                version_match = re.match(version_regex, line)
                release_match = re.match(release_regex, line)

                if version_match and not zstream and not release:
                    current_version = version_match.group(2)
                    if hasattr(self, '_use_version'):
                        updated_content = self._use_version
                    else:
                        updated_content = increase_version(current_version)

                    line = "".join([version_match.group(1), updated_content, "\n"])

                elif release_match:
                    current_release = release_match.group(2)
                    if hasattr(self, '_use_release'):
                        updated_content = self._use_release
                    elif release:
                        updated_content = increase_version(current_release)
                    elif zstream:
                        updated_content = increase_zstream(current_release)
                    else:
                        updated_content = reset_release(current_release)

                    line = "".join([release_match.group(1), updated_content, "\n"])

                out_f.write(line)

            in_f.close()
            out_f.close()
            shutil.move(self.spec_file + ".new", self.spec_file)

        new_version = get_spec_version_and_release(self.full_project_dir,
                self.spec_file_name)
        if new_version.strip() == "":
            msg = "Error getting bumped package version, try: \n"
            msg = msg + "  'rpm -q --specfile %s'" % self.spec_file
            error_out(msg)
        info_out("Tagging new version of %s: %s -> %s" % (self.project_name,
            old_version, new_version))
        return new_version

    def release_type(self):
        """ return short string which explain type of release.
            e.g. 'minor release
            Child classes probably want to override this.
        """
        return "release"

    def _update_package_metadata(self, new_version):
        """
        We track package metadata in the .tito/packages/ directory. Each
        file here stores the latest package version (for the git branch you
        are on) as well as the relative path to the project's code. (from the
        git root)
        """
        self._clear_package_metadata()

        new_version_w_suffix = self._get_suffixed_version(new_version)
        # Write out our package metadata:
        metadata_file = os.path.join(self.rel_eng_dir, "packages",
                self.project_name)

        with open(metadata_file, 'w') as f:
            f.write("%s %s\n" % (new_version_w_suffix, self.relative_project_dir))

        # Git add it (in case it's a new file):
        run_command("git add %s" % metadata_file)
        run_command("git add %s" % os.path.join(self.full_project_dir,
            self.spec_file_name))

        fmt = ('Automatic commit of package '
               '[%(name)s] %(release_type)s [%(version)s].')
        if self.config.has_option(BUILDCONFIG_SECTION, "tag_commit_message_format"):
            fmt = self.config.get(BUILDCONFIG_SECTION, "tag_commit_message_format")
        new_version_w_suffix = self._get_suffixed_version(new_version)
        try:
            msg = fmt % {
                'name': self.project_name,
                'release_type': self.release_type(),
                'version': new_version_w_suffix,
            }
        except KeyError:
            exc = sys.exc_info()[1]
            raise TitoException('Unknown placeholder %s in tag_commit_message_format'
                                % exc)

        run_command('git commit -m {0} -m {1} -m {2}'.format(
            quote(msg), quote("Created by command:"), quote(" ".join(sys.argv[:]))))

        new_tag = self._get_new_tag(new_version)
        tag_msg = "Tagging package [%s] version [%s] in directory [%s]." % \
                (self.project_name, new_tag,
                        self.relative_project_dir)

        # Optionally gpg sign the tag
        sign_tag = ""
        if self.config.has_option(BUILDCONFIG_SECTION, "sign_tag"):
            if self.config.getboolean(BUILDCONFIG_SECTION, "sign_tag"):
                sign_tag = "-s "

        run_command('git tag %s -m "%s" %s' % (sign_tag, tag_msg, new_tag))
        print
        info_out("Created tag: %s" % new_tag)
        print("   View: git show HEAD")
        print("   Undo: tito tag -u")
        print("   Push: git push --follow-tags origin")

    def _check_tag_does_not_exist(self, new_tag):
        status, output = getstatusoutput(
            'git tag -l %s|grep ""' % new_tag)
        if status == 0:
            raise Exception("Tag %s already exists!" % new_tag)

    def _clear_package_metadata(self):
        """
        Remove all .tito/packages/ files that have a relative path
        matching the package we're tagging a new version of. Normally
        this just removes the previous package file but if we were
        renaming oldpackage to newpackage, this would git rm
        .tito/packages/oldpackage and add
        .tito/packages/spacewalk-newpackage.
        """
        metadata_dir = os.path.join(self.rel_eng_dir, "packages")
        for filename in os.listdir(metadata_dir):
            metadata_file = os.path.join(metadata_dir, filename)  # full path

            if os.path.isdir(metadata_file) or filename.startswith("."):
                continue

            temp_file = open(metadata_file, 'r')
            (version, relative_dir) = temp_file.readline().split(" ")
            relative_dir = relative_dir.strip()  # sometimes has a newline

            if relative_dir == self.relative_project_dir:
                debug("Found metadata for our prefix: %s" %
                        metadata_file)
                debug("   version: %s" % version)
                debug("   dir: %s" % relative_dir)
                if filename == self.project_name:
                    debug("Updating %s with new version." %
                            metadata_file)
                else:
                    warn_out("%s also references %s" % (filename, self.relative_project_dir))
                    print("Assuming package has been renamed and removing it.")
                    run_command("git rm %s" % metadata_file)

    def _get_new_tag(self, version_and_release):
        """ Returns the actual tag we'll be creating. """
        suffixed_version = self._get_suffixed_version(self._get_version(version_and_release))
        release = self._get_release(version_and_release)
        return self._get_tag_for_version(suffixed_version, release)

    def _get_release(self, version_and_release):
        return version_and_release.split('-')[-1]

    def _get_version(self, version_and_release):
        return version_and_release.split('-')[-2]

    def _get_suffixed_version(self, version):
        """ If global config specifies a tag suffix, use it """
        suffix = ""
        if self.config.has_option(BUILDCONFIG_SECTION, "tag_suffix"):
            suffix = self.config.get(BUILDCONFIG_SECTION, "tag_suffix")
        return "{0}{1}".format(version, suffix)

    def _get_tag_for_version(self, version, release=''):
        """
        Determine what the tag will look like for a given version.
        Can be overridden when custom taggers override counterpart,
        tito.Builder._get_tag_for_version().
        """
        if self.config.has_option(BUILDCONFIG_SECTION, "tag_format"):
            tag_format = self.config.get(BUILDCONFIG_SECTION, "tag_format")
        else:
            tag_format = "{component}-{version}-{release}"
        kwargs = {
            'component': self.project_name,
            'version': version,
            'release': release
        }
        # Strip extra dashes if one of the params is empty
        return tag_format.format(**kwargs).strip('-')

    def _update_version_file(self, new_version):
        """
        land this new_version in the designated file
        and stages that file for a git commit
        """
        version_file = self._version_file_path()
        if not version_file:
            debug("No destination version file found, skipping.")
            return

        debug("Found version file to write: %s" % version_file)
        version_file_template = self._version_file_template()
        if version_file_template is None:
            error_out("Version file specified but without corresponding template.")

        t = Template(version_file_template)
        f = open(version_file, 'w')
        (new_ver, new_rel) = new_version.split('-')
        f.write(t.safe_substitute(
            version=new_ver,
            release=new_rel))
        f.close()

        run_command("git add %s" % version_file)

    def _version_file_template(self):
        """
        provide a configuration in tito.props to a file that is a
        python string.Template conforming blob, like
            [version_template]
            template_file = ./.tito/templates/my_java_properties

        variables defined inside the template are $version and $release

        see also http://docs.python.org/2/library/string.html#template-strings
        """
        if self.config.has_option("version_template", "template_file"):
            f = open(os.path.join(self.git_root,
                self.config.get("version_template", "template_file")), 'r')
            buf = f.read()
            f.close()
            return buf
        return None

    def _version_file_path(self):
        """
        provide a version file to write in tito.props, like
            [version_template]
            destination_file = ./foo.rb
        """
        if self.config.has_option("version_template", "destination_file"):
            return self.config.get("version_template", "destination_file")
        return None


class ReleaseTagger(VersionTagger):
    """
    Tagger which increments the spec file release instead of version.

    Used for:
      - Packages we build from a tarball checked directly into git.
      - Satellite packages built on top of Spacewalk tarballs.
    """

    def _tag_release(self):
        """
        Tag a new release of the package. (i.e. x.y.z-r+1)
        """
        self._make_changelog()
        # the user might have passed --use-version
        # so let's just bump the release if they did not
        bump_release = not hasattr(self, '_use_version')
        new_version = self._bump_version(release=bump_release)

        self._check_tag_does_not_exist(self._get_new_tag(new_version))
        self._update_changelog(new_version)
        self._update_package_metadata(new_version)

    def release_type(self):
        """ return short string "minor release" """
        return "minor release"


class ForceVersionTagger(VersionTagger):
    """
    Legacy Tagger which is chosen when a user passes the `--use-version`
    flag in on the command line. This implementation has been merged with
    the default `VersionTagger`, and remains only for backward compatibility
    for users that were overriding this class with a custom implementation.
    TODO: remove this class in a future release
    """



# Copyright (c) 2008-2011 Red Hat, Inc.
#
# This software is licensed to you under the GNU General Public License,
# version 2 (GPLv2). There is NO WARRANTY for this software, express or
# implied, including the implied warranties of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. You should have received a copy of GPLv2
# along with this software; if not, see
# http://www.gnu.org/licenses/old-licenses/gpl-2.0.txt.
#
# Red Hat trademarks are not licensed under GPLv2. No permission is
# granted to use or replicate Red Hat trademarks that are incorporated
# in this software or its documentation.
"""
Shared code for builder and tagger class
"""

import os
from tito.common import find_git_root, tito_config_dir


class ConfigObject(object):
    """
    Perent class for Builder and Tagger with shared code
    """

    def __init__(self, config=None):
        """
        config - Merged configuration. (global plus package specific)
        """
        self.config = config

        # Override global configurations using local configurations
        for section in config.sections():
            for options in config.options(section):
                if not self.config.has_section(section):
                    self.config.add_section(section)
                self.config.set(section, options,
                        config.get(section, options))

        self.git_root = find_git_root()
        self.rel_eng_dir = os.path.join(self.git_root, tito_config_dir())



import re
from tito.common import run_command
from tito.tagger import ReleaseTagger


class RHELTagger(ReleaseTagger):
    """
    Tagger which is based on ReleaseTagger and use Red Hat Enterprise Linux
    format of Changelog:
    - Resolves: #1111 - description
    or
    - Related: #1111 - description
    if BZ number was already mentioned in this changelog

    Used for:
        - Red Hat Enterprise Linux

    If you want it put in tito.pros:
    [buildconfig]
    tagger = tito.rheltagger.RHELTagger
    """

    def _generate_default_changelog(self, last_tag):
        """
        Run git-log and will generate changelog, which still can be edited by user
        in _make_changelog.
        use format:
        - Resolves: #1111 - description
        """
        patch_command = "git log --pretty='format:%%s%s'" \
                         " --relative %s..%s -- %s" % (self._changelog_format(), last_tag, "HEAD", ".")
        output = run_command(patch_command)
        BZ = {}
        result = None
        for line in reversed(output.split('\n')):
            line = self._changelog_remove_cherrypick(line)
            line = line.replace('%', '%%')

            # prepend Related/Resolves if subject contains BZ number
            m = re.match("(\d+)\s+-\s+(.*)", line)
            if m:
                bz_number = m.group(1)
                if bz_number in BZ:
                    line = "Related: #%s - %s" % (bz_number, m.group(2))
                else:
                    line = "Resolves: #%s - %s" % (bz_number, m.group(2))
                    BZ[bz_number] = 1
            if result:
                result = line + "\n" + result
            else:
                result = line
        return result




# Copyright (c) 2008-2014 Red Hat, Inc.
#
# This software is licensed to you under the GNU General Public License,
# version 2 (GPLv2). There is NO WARRANTY for this software, express or
# implied, including the implied warranties of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. You should have received a copy of GPLv2
# along with this software; if not, see
# http://www.gnu.org/licenses/old-licenses/gpl-2.0.txt.
#
# Red Hat trademarks are not licensed under GPLv2. No permission is
# granted to use or replicate Red Hat trademarks that are incorporated
# in this software or its documentation.
"""
Functional Tests for the YumReleaser.
"""

import glob
import os
import shutil
import tempfile

from os.path import join

from functional.fixture import TitoGitTestFixture, tito

from tito.compat import *  # NOQA
from tito.common import run_command


# There is not many simple options to check on what distribution this is running.
# Fortunately, we only need to check for Fedora Rawhide and EPEL6, so we can
# determine it from python version. This is compatible for all distributions.
import sys
is_rawhide = sys.version_info[:2] >= (3, 8)
is_epel6 = sys.version_info[:2] == (2, 6)

if is_epel6:
    import unittest2 as unittest
else:
    import unittest


PKG_NAME = "releaseme"

RELEASER_CONF = """
[yum-test]
releaser = tito.release.YumRepoReleaser
builder = tito.builder.Builder
srpm_disttag = .fc20
rsync = %s
"""


class YumReleaserTests(TitoGitTestFixture):

    def setUp(self):
        TitoGitTestFixture.setUp(self)
        self.create_project(PKG_NAME)

        # Setup test config:
        self.config = RawConfigParser()
        self.config.add_section("buildconfig")
        self.config.set("buildconfig", "builder",
                "tito.builder.Builder")
        self.config.set("buildconfig", "offline",
                "true")

        self.output_dir = tempfile.mkdtemp("-titotestoutput")

    def tearDown(self):
        TitoGitTestFixture.tearDown(self)
        shutil.rmtree(self.output_dir)
        pass

    def _setup_fetchbuilder_releaser(self, yum_repo_dir):
        self.write_file(join(self.repo_dir, '.tito/releasers.conf'),
                RELEASER_CONF % yum_repo_dir)

    # createrepo_c (0.15.5+ which is in rawhide) currently coredumps
    # https://github.com/rpm-software-management/createrepo_c/issues/202
    @unittest.skipIf(is_rawhide, "Re-enable once createrepo_c #202 gets fixed")
    def test_with_releaser(self):
        yum_repo_dir = os.path.join(self.output_dir, 'yum')
        run_command('mkdir -p %s' % yum_repo_dir)
        self._setup_fetchbuilder_releaser(yum_repo_dir)
        tito('release --debug yum-test')

        self.assertEquals(1, len(glob.glob(join(yum_repo_dir,
            "releaseme-0.0.1-1.*noarch.rpm"))))
        self.assertEquals(1, len(glob.glob(join(yum_repo_dir,
            "repodata/repomd.xml"))))



#
# Copyright (c) 2008-2014 Red Hat, Inc.
#
# This software is licensed to you under the GNU General Public License,
# version 2 (GPLv2). There is NO WARRANTY for this software, express or
# implied, including the implied warranties of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. You should have received a copy of GPLv2
# along with this software; if not, see
# http://www.gnu.org/licenses/old-licenses/gpl-2.0.txt.
#
# Red Hat trademarks are not licensed under GPLv2. No permission is
# granted to use or replicate Red Hat trademarks that are incorporated
# in this software or its documentation.

import os
import tempfile
from tito.builder import Builder
from tito.common import run_command
from tito.compat import RawConfigParser
from functional.fixture import TitoGitTestFixture, tito

PKG_NAME = "titotestpkg"


class BuilderTests(TitoGitTestFixture):

    def setUp(self):
        TitoGitTestFixture.setUp(self)
        os.chdir(self.repo_dir)

        self.config = RawConfigParser()
        self.output_dir = tempfile.mkdtemp("-titotestoutput")

    def test_scl_from_options(self):
        self.create_project(PKG_NAME)
        builder = Builder(PKG_NAME, None, self.output_dir,
            self.config, {}, {'scl': ['ruby193']}, **{'offline': True})
        self.assertEqual('ruby193', builder.scl)

    def test_scl_from_kwargs(self):
        self.create_project(PKG_NAME)
        builder = Builder(PKG_NAME, None, self.output_dir,
            self.config, {}, {}, **{'offline': True, 'scl': 'ruby193'})
        self.assertEqual('ruby193', builder.scl)

    def test_rpmbuild_options_from_options(self):
        self.create_project(PKG_NAME)
        builder = Builder(PKG_NAME, None, self.output_dir,
            self.config, {}, {'rpmbuild_options': ['--define "foo bar"',
            '--define "bar baz"']}, **{'offline': True})
        self.assertEqual('--define "foo bar" --define "bar baz"', builder.rpmbuild_options)

    def test_rpmbuild_options_from_kwargs(self):
        self.create_project(PKG_NAME)
        builder = Builder(PKG_NAME, None, self.output_dir,
            self.config, {}, {}, **{'offline': True, 'rpmbuild_options': '--define "foo bar"'})
        self.assertEqual('--define "foo bar"', builder.rpmbuild_options)

    def test_rpmbuild_options_missing(self):
        self.create_project(PKG_NAME)
        builder = Builder(PKG_NAME, None, self.output_dir,
            self.config, {}, {}, **{'offline': True})
        self.assertEqual('', builder.rpmbuild_options)

    def test_untagged_test_version(self):
        self.create_project(PKG_NAME, tag=False)
        self.assertEqual("", run_command("git tag -l").strip())

        builder = Builder(PKG_NAME, None, self.output_dir,
            self.config, {}, {}, **{'offline': True, 'test': True})
        self.assertEqual('0.0.1-1', builder.build_version)

    def test_untagged_test_build(self):
        self.create_project(PKG_NAME, tag=False)
        self.assertEqual("", run_command("git tag -l").strip())
        tito('build --srpm --test')



# Copyright (c) 2008-2014 Red Hat, Inc.
#
# This software is licensed to you under the GNU General Public License,
# version 2 (GPLv2). There is NO WARRANTY for this software, express or
# implied, including the implied warranties of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. You should have received a copy of GPLv2
# along with this software; if not, see
# http://www.gnu.org/licenses/old-licenses/gpl-2.0.txt.
#
# Red Hat trademarks are not licensed under GPLv2. No permission is
# granted to use or replicate Red Hat trademarks that are incorporated
# in this software or its documentation.

import re
import os.path
import shutil

from tito.builder.main import BuilderBase
from tito.config_object import ConfigObject
from tito.common import error_out, debug, get_spec_version_and_release, \
    get_class_by_name


class FetchBuilder(ConfigObject, BuilderBase):
    """
    A separate Builder class for projects whose source is not in git. Source
    is fetched via a configurable strategy, which also determines what version
    and release to insert into the spec file.

    Cannot build past tags.
    """
    # TODO: test only for now, setup a tagger to fetch sources and store in git annex,
    # then we can do official builds as well.
    REQUIRED_ARGS = []

    def __init__(self, name=None, tag=None, build_dir=None,
            config=None, user_config=None,
            args=None, **kwargs):

        BuilderBase.__init__(self, name=name, build_dir=build_dir,
                config=config,
                user_config=user_config, args=args, **kwargs)

        if tag:
            error_out("FetchBuilder does not support building specific tags.")

        if not config.has_option("builder",
                "fetch_strategy"):
            print("WARNING: no fetch_strategy specified in tito.props"
                    ", assuming ArgSourceStrategy.")
            if not config.has_section("builder"):
                config.add_section("builder")
            config.set('builder', 'fetch_strategy',
                    'tito.builder.fetch.ArgSourceStrategy')

        self.build_tag = '%s-%s' % (self.project_name,
                get_spec_version_and_release(self.start_dir, '%s.spec' % self.project_name))

    def tgz(self):
        self.ran_tgz = True
        self._create_build_dirs()

        print("Fetching sources...")
        source_strat_class = get_class_by_name(self.config.get(
            'builder', 'fetch_strategy'))
        source_strat = source_strat_class(self)
        source_strat.fetch()
        self.sources = source_strat.sources
        self.spec_file = source_strat.spec_file

    def _get_rpmbuild_dir_options(self):
        return ('--define "_topdir %s" --define "_sourcedir %s" --define "_builddir %s" '
            '--define "_srcrpmdir %s" --define "_rpmdir %s" ' % (
                self.rpmbuild_dir,
                self.rpmbuild_sourcedir, self.rpmbuild_builddir,
                self.rpmbuild_basedir, self.rpmbuild_basedir))


class SourceStrategy(object):
    """
    Base class for source strategies. These are responsible for fetching the
    sources the builder will use, and determining what version/release we're
    building.

    This is created and run in the tgz step of the builder. It will be passed
    a reference to the builder calling it, which will be important for accessing
    a lot of required information.

    Ideally sources and the spec file to be used should be copied into
    builder.rpmbuild_sourcedir, which will be cleaned up automatically after
    the builder runs.
    """
    def __init__(self, builder):
        """
        Defines fields that should be set when a sub-class runs fetch.
        """
        self.builder = builder

        # Full path to the spec file we'll actually use to build, should be a
        # copy, never a live spec file in a git repo as sometimes it will be
        # modified:
        self.spec_file = None

        # Will contain the full path to each source we gather:
        self.sources = []

        # The version we're building:
        self.version = None

        # The release we're building:
        self.release = None

    def fetch(self):
        raise NotImplementedError()


class ArgSourceStrategy(SourceStrategy):
    """
    Assumes the builder was passed an explicit argument specifying which source
    file(s) to use.
    """
    def fetch(self):

        # Assuming we're still in the start directory, get the absolute path
        # to all sources specified:
        # TODO: support passing of multiple sources here.
        # TODO: error out if not present
        manual_sources = [self.builder.args['source'][0]]
        debug("Got sources: %s" % manual_sources)

        # Copy the live spec from our starting location. Unlike most builders,
        # we are not using a copy from a past git commit.
        self.spec_file = os.path.join(self.builder.rpmbuild_sourcedir,
                    '%s.spec' % self.builder.project_name)
        shutil.copyfile(
            os.path.join(self.builder.start_dir, '%s.spec' %
                self.builder.project_name),
            self.spec_file)
        print("  %s.spec" % self.builder.project_name)

        # TODO: Make this a configurable strategy:
        i = 0
        replacements = []
        for s in manual_sources:
            base_name = os.path.basename(s)
            dest_filepath = os.path.join(self.builder.rpmbuild_sourcedir,
                    base_name)
            shutil.copyfile(s, dest_filepath)
            self.sources.append(dest_filepath)

            # Add a line to replace in the spec for each source:
            source_regex = re.compile("^(source%s:\s*)(.+)$" % i, re.IGNORECASE)
            new_line = "Source%s: %s\n" % (i, base_name)
            replacements.append((source_regex, new_line))

        # Replace version and release in spec:
        version_regex = re.compile("^(version:\s*)(.+)$", re.IGNORECASE)
        release_regex = re.compile("^(release:\s*)(.+)$", re.IGNORECASE)

        (self.version, self.release) = self._get_version_and_release()
        print("Building version: %s" % self.version)
        print("Building release: %s" % self.release)
        replacements.append((version_regex, "Version: %s\n" % self.version))
        replacements.append((release_regex, "Release: %s\n" % self.release))

        self.replace_in_spec(replacements)

    def _get_version_and_release(self):
        """
        Get the version and release from the builder.
        Sources are configured at this point.
        """
        # Assuming source0 is a tar.gz we can extract a version and possibly
        # release from:
        base_name = os.path.basename(self.sources[0])
        debug("Extracting version/release from: %s" % base_name)

        # usually a source tarball won't have a release, that is an RPM concept.
        # Don't forget dist!
        release = "1%{?dist}"

        # Example filename: tito-0.4.18.tar.gz:
        simple_version_re = re.compile(".*-(.*).(tar.gz|tgz|zip|bz2)")
        match = re.search(simple_version_re, base_name)
        if match:
            version = match.group(1)
        else:
            error_out("Unable to determine version from file: %s" % base_name)

        return (version, release)

    def replace_in_spec(self, replacements):
        """
        Replace lines in the spec file using the given replacements.

        Replacements are a tuple of a regex to look for, and a new line to
        substitute in when the regex matches.

        Replaces all lines with one pass through the file.
        """
        in_f = open(self.spec_file, 'r')
        out_f = open(self.spec_file + ".new", 'w')
        for line in in_f.readlines():
            for line_regex, new_line in replacements:
                match = re.match(line_regex, line)
                if match:
                    line = new_line
            out_f.write(line)

        in_f.close()
        out_f.close()
        shutil.move(self.spec_file + ".new", self.spec_file)



# Copyright (c) 2013 Red Hat, Inc.
#
# This software is licensed to you under the GNU General Public License,
# version 2 (GPLv2). There is NO WARRANTY for this software, express or
# implied, including the implied warranties of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. You should have received a copy of GPLv2
# along with this software; if not, see
# http://www.gnu.org/licenses/old-licenses/gpl-2.0.txt.
#
# Red Hat trademarks are not licensed under GPLv2. No permission is
# granted to use or replicate Red Hat trademarks that are incorporated
# in this software or its documentation.

import os.path
import subprocess

from tito.common import run_command, info_out, error_out
from tito.release import KojiReleaser


class CoprReleaser(KojiReleaser):
    """ Releaser for Copr using copr-cli command """

    REQUIRED_CONFIG = ['project_name']
    cli_tool = "copr-cli"
    NAME = "Copr"

    def __init__(self, name=None, tag=None, build_dir=None,
            config=None, user_config=None,
            target=None, releaser_config=None, no_cleanup=False,
            test=False, auto_accept=False, **kwargs):
        self.user_config = user_config

        KojiReleaser.__init__(self, name, tag, build_dir, config,
                user_config, target, releaser_config, no_cleanup, test,
                auto_accept, **kwargs)

        self.copr_project_name = \
            self.releaser_config.get(self.target, "project_name")
        self.srpm_submitted = False

        # 'copr-cli build' options, e.g. --chroot
        # Default to --nowait, so that it mirrors the behavior of fedpkg
        self.copr_options = '--nowait'
        if self.releaser_config.has_option(self.target, "copr_options"):
            self.copr_options = \
                self.releaser_config.get(self.target, "copr_options")

    def autobuild_tags(self):
        """ will return list of project for which we are building """
        result = self.releaser_config.get(self.target, "project_name")
        return result.strip().split(" ")

    def _koji_release(self):
        self.srpm_submitted = False
        if not self.builder.config.has_section(self.copr_project_name):
            self.builder.config.add_section(self.copr_project_name)
        KojiReleaser._koji_release(self)

    def _check_releaser_config(self):
        """
        Verify this release target has all the config options it needs.
        """
        self.remote_location = None
        if self.releaser_config.has_option(self.target, "remote_location"):
            self.remote_location = self.releaser_config.get(self.target, "remote_location")
        elif 'COPR_REMOTE_LOCATION' in self.user_config:
            self.remote_location = self.user_config['COPR_REMOTE_LOCATION']
        KojiReleaser._check_releaser_config(self)

    def _submit_build(self, executable, koji_opts, tag, srpm_location):
        """
        Submit the build into Copr
        If `remote_location` was specified, the SRPM package is uploaded
        there and then submitted into Copr via URL. Otherwise the package
        is submitted from local storage directly to the Copr
        """
        path = srpm_location
        if self.remote_location:
            path = self.remote_location + os.path.basename(srpm_location)
            self._upload(srpm_location)
        self._submit(path, tag)

    def _upload(self, srpm_location):
        if self.srpm_submitted:
            srpm_location = self.srpm_submitted

        # e.g. "scp %(srpm)s my.web.com:public_html/my_srpm/"
        cmd = self.releaser_config.get(self.target, "upload_command")
        cmd_upload = cmd % {'srpm': srpm_location}

        if self.dry_run:
            self.print_dry_run_warning(cmd_upload)
            return

        # TODO: no error handling when run_command fails:
        if not self.srpm_submitted:
            print("Uploading src.rpm.")
            print(run_command(cmd_upload))
            self.srpm_submitted = srpm_location

    def _submit(self, srpm_location, project):
        cmd_submit = "/usr/bin/%s build %s %s %s" % \
                     (self.cli_tool, self.copr_options, project, srpm_location)
        if self.dry_run:
            self.print_dry_run_warning(cmd_submit)
            return

        info_out("Submitting build into %s." % self.NAME)
        self._run_command(cmd_submit)

    def _run_command(self, cmd):
        process = subprocess.Popen(cmd.split())
        process.wait()
        if process.returncode > 0:
            error_out("Failed running `%s`" % cmd)



# Copyright (c) 2008-2010 Red Hat, Inc.
#
# This software is licensed to you under the GNU General Public License,
# version 2 (GPLv2). There is NO WARRANTY for this software, express or
# implied, including the implied warranties of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. You should have received a copy of GPLv2
# along with this software; if not, see
# http://www.gnu.org/licenses/old-licenses/gpl-2.0.txt.
#
# Red Hat trademarks are not licensed under GPLv2. No permission is
# granted to use or replicate Red Hat trademarks that are incorporated
# in this software or its documentation.

# flake8: noqa

"""
Compatibility library for Python 2.4 up through Python 3.
"""
import os
import sys
import contextlib
ENCODING = sys.getdefaultencoding()
PY2 = sys.version_info[0] == 2
if PY2:
    import commands
    from ConfigParser import NoOptionError
    from ConfigParser import RawConfigParser
    from StringIO import StringIO
    from urlparse import urlparse
    import xmlrpclib
    text_type = unicode
    binary_type = str
else:
    import subprocess
    from configparser import NoOptionError
    from configparser import RawConfigParser
    from io import StringIO
    from urllib.parse import urlparse
    from contextlib import redirect_stdout
    import xmlrpc.client as xmlrpclib
    text_type = str
    binary_type = bytes


def ensure_text(x, encoding="utf8"):
    if isinstance(x, binary_type):
        return x.decode(encoding)
    elif isinstance(x, text_type):
        return x
    else:
        raise TypeError("Not expecting type '%s'" % type(x))


def ensure_binary(x, encoding="utf8"):
    if isinstance(x, text_type):
        return x.encode(encoding)
    elif isinstance(x, binary_type):
        return x
    else:
        raise TypeError("Not expecting type '%s'" % type(x))


def getstatusoutput(cmd):
    """
    Returns (status, output) of executing cmd in a shell.
    Supports Python 2.4 and 3.x.
    """
    if PY2:
        return commands.getstatusoutput(cmd)
    else:
        return subprocess.getstatusoutput(cmd)


def getoutput(cmd):
    """
    Returns output of executing cmd in a shell.
    Supports Python 2.4 and 3.x.
    """
    return getstatusoutput(cmd)[1]


def dictionary_override(d1, d2):
    """
    Return a new dictionary object where
    d2 elements override d1 elements.
    """
    if PY2:
        overrides = d1.items() + d2.items()
    else:
        overrides = list(d1.items()) + list(d2.items())
    return dict(overrides)


def write(fd, str):
    """
    A version of os.write that
    supports Python 2.4 and 3.x.
    """
    if PY2:
        os.write(fd, str)
    else:
        os.write(fd, bytes(str, ENCODING))


if PY2:
    @contextlib.contextmanager
    def redirect_stdout(target):
        original = sys.stdout
        try:
            sys.stdout = target
            yield
        finally:
            sys.stdout = original



#
# Copyright (c) 2008-2014 Red Hat, Inc.
#
# This software is licensed to you under the GNU General Public License,
# version 2 (GPLv2). There is NO WARRANTY for this software, express or
# implied, including the implied warranties of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. You should have received a copy of GPLv2
# along with this software; if not, see
# http://www.gnu.org/licenses/old-licenses/gpl-2.0.txt.
#
# Red Hat trademarks are not licensed under GPLv2. No permission is
# granted to use or replicate Red Hat trademarks that are incorporated
# in this software or its documentation.

"""
Use pep8 to check for errors or deprecations that can cause Python 3 to fail.
source: https://github.com/jcrocholl/pep8
docs:   http://pep8.readthedocs.org/en/latest/intro.html

Python 3 is picky about indentation:
http://docs.python.org/3.3/reference/lexical_analysis.html
"""

try:
    # python-pep8 package is retired in Fedora because upstream
    # moved to pycodestyle. Please see
    # https://bugzilla.redhat.com/show_bug.cgi?id=1667200
    import pep8
except ImportError:
    import pycodestyle as pep8


from tito.compat import *  # NOQA
from tito.compat import StringIO, redirect_stdout
from unit.fixture import TitoUnitTestFixture, REPO_DIR


class TestPep8(TitoUnitTestFixture):
    def setUp(self):
        TitoUnitTestFixture.setUp(self)

    def test_conformance(self):
        tests = [
            # http://pep8.readthedocs.org/en/latest/intro.html#error-codes
            'E101',  # indentation contains mixed spaces and tabs
            'E111',  # indentation is not a multiple of four
            'E112',  # expected an indented block
            'E113',  # unexpected indentation
            'E121',  # continuation line indentation is not a multiple of four
            'E122',  # continuation line missing indentation or outdented
            'E126',  # continuation line over-indented for hanging indent
            'E2',    # whitespace errors
            'E3',    # blank line errors
            'E4',    # import errors
            'E502',  # the backslash is redundant between brackets
            'E9',    # runtime errors (SyntaxError, IndentationError, IOError)
            'W1',    # indentation warnings
            'W2',    # whitespace warnings
            'W3',    # blank line warnings

            # @FIXME we currently have a lot of these errors introduced to our
            # codebase. Let's temporarily disable the check, so we can get travis
            # working again.
            # 'E7',    # statement errors
            # 'W6',    # deprecated features
        ]

        try:
            checker = pep8.StyleGuide(select=tests, paths=[REPO_DIR], reporter=pep8.StandardReport)
            # Unfortunatelly, IMHO the most interesting information
            # `checker.check_files()` prints to the STDOUT and doesn't provide
            # API to get it - the list of _bad_ files. Let's just read it from
            # the output
            with StringIO() as buf, redirect_stdout(buf):
                report = checker.check_files()
                result = report.total_errors
                output = buf.getvalue()

        except AttributeError:
            # We don't have pep8.StyleGuide, so we must be
            # using pep8 older than git tag 1.1-72-gf20d656.
            os.chdir(REPO_DIR)
            checks = ','.join(tests)
            cmd = "pep8 --select=%s %s | wc -l" % (checks, '.')
            result = int(getoutput(cmd))
            output = getoutput("pep8 --select %s %s" % (checks, '.'))

        if result != 0:
            self.fail("Found PEP8 errors that may break your code in Python 3:\n%s" % output)


class UglyHackishTest(TitoUnitTestFixture):
    def setUp(self):
        TitoUnitTestFixture.setUp(self)
        os.chdir(REPO_DIR)

    def test_exceptions_3(self):
        # detect 'except rpm.error, e:'
        regex = "'^[[:space:]]*except [^,]+,[[:space:]]*[[:alpha:]]+:'"
        cmd = "find . -type f -regex '.*\.py$' -exec egrep %s {} + | wc -l" % regex
        result = int(getoutput(cmd))
        self.assertEqual(result, 0, "Found except clause not supported in Python 3")

    def test_import_commands(self):
        cmd = "find . -type f -regex '.*\.py$' -exec egrep '^(import|from) commands\.' {} + | grep -v 'compat\.py' | wc -l"
        result = int(getoutput(cmd))
        self.assertEqual(result, 0, "Found commands module (not supported in Python 3)")

    def test_use_commands(self):
        cmd = "find . -type f -regex '.*\.py$' -exec egrep 'commands\.' {} + | grep -v 'compat\.py' | wc -l"
        result = int(getoutput(cmd))
        self.assertEqual(result, 0, "Found commands module (not supported in Python 3)")

    def test_print_function(self):
        cmd = "find . -type f -regex '.*\.py$' -exec grep '^[[:space:]]*print .*' {} + | wc -l"
        result = int(getoutput(cmd))
        self.assertEqual(result, 0, "Found print statement (not supported in Python 3)")



#
# Copyright (c) 2008-2010 Red Hat, Inc.
#
# This software is licensed to you under the GNU General Public License,
# version 2 (GPLv2). There is NO WARRANTY for this software, express or
# implied, including the implied warranties of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. You should have received a copy of GPLv2
# along with this software; if not, see
# http://www.gnu.org/licenses/old-licenses/gpl-2.0.txt.
#
# Red Hat trademarks are not licensed under GPLv2. No permission is
# granted to use or replicate Red Hat trademarks that are incorporated
# in this software or its documentation.
"""
Functional Tests for Tito at the CLI level.

NOTE: These tests require a makeshift git repository created in /tmp.
"""

import os
from os.path import join

from tito.common import run_command, \
    get_latest_tagged_version, tag_exists_locally
from functional.fixture import TitoGitTestFixture, tito

# A location where we can safely create a test git repository.
# WARNING: This location will be destroyed if present.

TEST_PKG_1 = 'titotestpkg'

TEST_PKG_2 = 'titotestpkg2'

TEST_PKG_3 = 'titotestpkg3'

TEST_PKGS = [TEST_PKG_1, TEST_PKG_2, TEST_PKG_3]


def release_bumped(initial_version, new_version):
    first_release = initial_version.split('-')[-1]
    new_release = new_version.split('-')[-1]
    return new_release == str(int(first_release) + 1)


TEMPLATE_TAGGER_TITO_PROPS = """
[buildconfig]
tagger = tito.tagger.VersionTagger
builder = tito.builder.Builder

[version_template]
destination_file = version.txt
template_file = .tito/templates/version.rb
"""

VERSION_TEMPLATE_FILE = """
module Iteng
    module Util
      VERSION = "$version-$release"
    end
  end
"""


class MultiProjectTests(TitoGitTestFixture):

    def setUp(self):
        TitoGitTestFixture.setUp(self)

        self.create_project(TEST_PKG_1, os.path.join(self.repo_dir, 'pkg1'))
        self.create_project(TEST_PKG_2, os.path.join(self.repo_dir, 'pkg2'))
        self.create_project(TEST_PKG_3, os.path.join(self.repo_dir, 'pkg3'))

        # For second test package, use a tito.props to override and use the
        # ReleaseTagger:
        filename = os.path.join(self.repo_dir, 'pkg2', "tito.props")
        out_f = open(filename, 'w')
        out_f.write("[buildconfig]\n")
        out_f.write("tagger = tito.tagger.ReleaseTagger\n")
        out_f.write("builder = tito.builder.Builder\n")
        out_f.close()

        os.chdir(self.repo_dir)
        run_command('git add pkg2/tito.props')
        run_command("git commit -m 'add tito.props for pkg2'")

    def test_template_version_tagger(self):
        """
        Make sure the template is applied and results in the correct file
        being included in the tag.
        """
        pkg_dir = join(self.repo_dir, 'pkg3')
        filename = join(pkg_dir, "tito.props")
        self.write_file(filename, TEMPLATE_TAGGER_TITO_PROPS)
        run_command('mkdir -p %s' % join(self.repo_dir, '.tito/templates'))
        self.write_file(join(self.repo_dir,
            '.tito/templates/version.rb'), VERSION_TEMPLATE_FILE)

        os.chdir(self.repo_dir)
        run_command('git add pkg3/tito.props')
        run_command("git commit -m 'add tito.props for pkg3'")

        # Create another pkg3 tag and make sure we got a generated
        # template file.
        os.chdir(os.path.join(self.repo_dir, 'pkg3'))
        tito('tag --debug --accept-auto-changelog')
        new_ver = get_latest_tagged_version(TEST_PKG_3)
        self.assertEquals("0.0.2-1", new_ver)

        dest_file = os.path.join(self.repo_dir, 'pkg3', "version.txt")
        self.assertTrue(os.path.exists(dest_file))

        f = open(dest_file, 'r')
        contents = f.read()
        f.close()

        self.assertTrue("VERSION = \"0.0.2-1\"" in contents)

    def test_initial_tag_keep_version(self):
        # Tags were actually created in setup code:
        for pkg_name in TEST_PKGS:
            self.assertTrue(tag_exists_locally("%s-0.0.1-1" % pkg_name))
            self.assertTrue(os.path.exists(os.path.join(self.repo_dir,
                ".tito/packages", pkg_name)))

    def test_release_tagger(self):
        os.chdir(os.path.join(self.repo_dir, 'pkg2'))
        start_ver = get_latest_tagged_version(TEST_PKG_2)
        tito('tag --debug --accept-auto-changelog')
        new_ver = get_latest_tagged_version(TEST_PKG_2)
        self.assertTrue(release_bumped(start_ver, new_ver))

    def test_release_tagger_use_release(self):
        os.chdir(os.path.join(self.repo_dir, 'pkg2'))
        tito('tag --debug --accept-auto-changelog --use-release 42')
        new_ver = get_latest_tagged_version(TEST_PKG_2)
        self.assertEquals(new_ver.split('-')[-1], "42")

    def test_release_tagger_use_version(self):
        os.chdir(os.path.join(self.repo_dir, 'pkg2'))
        start_ver = get_latest_tagged_version(TEST_PKG_2)
        tito('tag --debug --accept-auto-changelog --use-version 1.3.37')
        new_ver = get_latest_tagged_version(TEST_PKG_2)
        self.assertFalse(release_bumped(start_ver, new_ver))
        self.assertEquals(new_ver, "1.3.37-1")

    def test_build_tgz(self):
        os.chdir(os.path.join(self.repo_dir, 'pkg1'))
        artifacts = tito('build --tgz')
        self.assertEquals(1, len(artifacts))
        self.assertEquals('%s-0.0.1.tar.gz' % TEST_PKG_1,
                os.path.basename(artifacts[0]))

    def test_build_rpm(self):
        os.chdir(os.path.join(self.repo_dir, 'pkg1'))
        artifacts = tito('build --rpm')
        self.assertEquals(3, len(artifacts))



# Copyright (c) 2012-2016 SUSE Linux Products GmbH
# Copyright (c) 2018 Neal Gompa.
#
# This software is licensed to you under the GNU General Public License,
# version 2 (GPLv2). There is NO WARRANTY for this software, express or
# implied, including the implied warranties of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. You should have received a copy of GPLv2
# along with this software; if not, see
# http://www.gnu.org/licenses/old-licenses/gpl-2.0.txt.
"""
Code for tagging packages in SUSE Style.
"""
import os
import re
import shutil
import subprocess
import tempfile
import textwrap
from tito.common import (debug, run_command, get_latest_tagged_version)
from tito.compat import write, StringIO
from tito.tagger import VersionTagger

from time import strftime


class SUSETagger(VersionTagger):
    """
    Tagger which is based on VersionTagger and use SUSE format of Changelog
    and SUSE specific changes file:

    If you want it put in tito.pros (global) or localy in build.py.props:
    [buildconfig]
    tagger = tito.susetagger.SUSETagger
    """

    def __init__(self, config=None, keep_version=False, offline=False, user_config=None):
        VersionTagger.__init__(self, config=config, keep_version=keep_version,
                               offline=offline, user_config=user_config)
        self.today = strftime("%a %b %d %T %Z %Y")
        self.changes_file_name = self.spec_file_name.replace('.spec', '.changes')
        self.changes_file = os.path.join(self.full_project_dir,
                self.changes_file_name)
        self._new_changelog_msg = "Initial package release"
        self.changelog_regex = re.compile('^\s%s\s%s(\s<%s>)?' % (self.today,
            self.git_user, self.git_email.replace("+", "\+").replace(".", "\.")))

    def _make_changelog(self):
        """
        Create a new changelog entry in the changes, with line items from git
        """
        if self._no_auto_changelog:
            debug("Skipping changelog generation.")
            return

        # Attempt to open the file if it exists, but create it if it doesn't
        try:
            in_f = open(self.changes_file, 'r')
        except FileNotFoundError:
            new_in_f = open(self.changes_file, 'w')
            new_in_f.close()
            in_f = open(self.changes_file, 'r')

        out_f = open(self.changes_file + ".new", 'w')

        old_version = get_latest_tagged_version(self.project_name)

        # don't die if this is a new package with no history
        if old_version is not None:
            last_tag = "%s-%s" % (self.project_name, old_version)
            output = self._generate_default_changelog(last_tag)
        else:
            output = self._new_changelog_msg

        fd, name = tempfile.mkstemp()
        write(fd, "# Create your changelog entry below:\n")
        header = "-------------------------------------------------------------------\n"
        if self.git_email is None or (('HIDE_EMAIL' in self.user_config) and
                (self.user_config['HIDE_EMAIL'] not in ['0', ''])):
            header = header + "%s - %s\n\n" % (self.today, self.git_user)
        else:
            header = header + "%s - %s <%s>\n\n" % (self.today, self.git_user,
                                                        self.git_email)

        write(fd, header)

        for cmd_out in output.split("\n"):
            write(fd, "- ")
            write(fd, "\n  ".join(textwrap.wrap(cmd_out, 77)))
            write(fd, "\n")

        write(fd, "\n")

        if not self._accept_auto_changelog:
            write(fd, "###################################################\n")
            write(fd, "# These are the already existing changelog entries:\n")
            write(fd, "###################################################\n")
            for line in in_f.readlines():
                write(fd, "#" + line)
            in_f.seek(0, 0)

            # Give the user a chance to edit the generated changelog:
            editor = 'vi'
            if "EDITOR" in os.environ:
                editor = os.environ["EDITOR"]
            subprocess.call([editor, name])

        os.lseek(fd, 0, 0)
        file = os.fdopen(fd)

        for line in file.readlines():
            if not line.startswith("#"):
                out_f.write(line)

        output = file.read()

        file.close()
        os.unlink(name)

        for line in in_f.readlines():
            out_f.write(line)

        in_f.close()
        out_f.close()

        shutil.move(self.changes_file + ".new", self.changes_file)

    def _update_changelog(self, new_version):
        """
        Update the changelog with the new version.
        """
        # Not thrilled about having to re-read the file here but we need to
        # check for the changelog entry before making any modifications, then
        # bump the version, then update the changelog.
        f = open(self.changes_file, 'r')
        buf = StringIO()
        found_match = False
        done = False
        empty_line_regex = re.compile('^\s*$')

        for line in f.readlines():
            if not done and not found_match and self.changelog_regex.match(line):
                buf.write(line)
                found_match = True
            elif not done and found_match and empty_line_regex.match(line):
                buf.write("\n- version %s\n" % new_version)
                done = True
            else:
                buf.write(line)
        f.close()

        # Write out the new file contents with our modified changelog entry:
        f = open(self.changes_file, 'w')
        f.write(buf.getvalue())
        f.close()
        buf.close()

    def _update_package_metadata(self, new_version):
        """
        We track package metadata in the rel-eng/packages/ directory. Each
        file here stores the latest package version (for the git branch you
        are on) as well as the relative path to the project's code. (from the
        git root)
        """
        self._clear_package_metadata()

        suffix = ""
        # If global config specifies a tag suffix, use it:
        if self.config.has_option("globalconfig", "tag_suffix"):
            suffix = self.config.get("globalconfig", "tag_suffix")

        new_version_w_suffix = "%s%s" % (new_version, suffix)
        # Write out our package metadata:
        metadata_file = os.path.join(self.rel_eng_dir, "packages",
                self.project_name)
        f = open(metadata_file, 'w')
        f.write("%s %s\n" % (new_version_w_suffix, self.relative_project_dir))
        f.close()

        # Git add it (in case it's a new file):
        run_command("git add %s" % metadata_file)
        run_command("git add %s" % os.path.join(self.full_project_dir,
            self.spec_file_name))
        if not self._no_auto_changelog:
            run_command("git add %s" % os.path.join(self.full_project_dir,
                self.changes_file_name))

        run_command('git commit -m "Automatic commit of package ' +
                '[%s] %s [%s]."' % (self.project_name, self.release_type(),
                    new_version_w_suffix))

        tag_msg = "Tagging package [%s] version [%s] in directory [%s]." % \
                (self.project_name, new_version_w_suffix,
                        self.relative_project_dir)

        new_tag = self._get_new_tag(new_version)
        run_command('git tag -m "%s" %s' % (tag_msg, new_tag))
        print
        print("Created tag: %s" % new_tag)
        print("   View: git show HEAD")
        print("   Undo: tito tag -u")
        print("   Push: git push origin HEAD && git push origin %s" % new_tag)
        print("or Push: git push origin HEAD && git push origin --tags")



import unittest
import mock
from tito.compat import PY2, RawConfigParser
from tito.release import Releaser
from unit import builtins_input


class ReleaserTests(unittest.TestCase):

    @mock.patch("tito.release.main.create_builder")
    @mock.patch("tito.release.main.mkdtemp")
    def setUp(self, mkdtemp, create_builder):
        self.config = RawConfigParser()

        self.releaser_config = RawConfigParser()
        self.releaser_config.add_section("test")
        self.releaser_config.set('test', "releaser",
            "tito.release.Releaser")

        self.releaser = Releaser("titotestpkg", None, "/tmp/tito/",
            self.config, {}, "test", self.releaser_config, False,
            False, False, **{"offline": True})

    @mock.patch(builtins_input)
    def test_ask_yes_or_no(self, input_mock):
        input_mock.side_effect = "y"
        assert self.releaser._ask_yes_no()

        input_mock.side_effect = "n"
        assert not self.releaser._ask_yes_no()

        input_mock.side_effect = ["yy", "y"]
        assert self.releaser._ask_yes_no()

        input_mock.side_effect = ["yy", "no"]
        assert not self.releaser._ask_yes_no()



import unittest
import mock
from datetime import datetime
from tito.tagger import VersionTagger
from tito.compat import PY2, RawConfigParser

from unit import is_epel6
if is_epel6:
    import unittest2 as unittest
else:
    import unittest


def strftime_mock(s):
    if not PY2:
        from datetime import timezone
    tzinfo = None if PY2 else timezone.utc
    someday = datetime(2020, 4, 22, 15, 2, tzinfo=tzinfo)
    return someday.strftime(s)


class TestVersionTagger(unittest.TestCase):
    def setUp(self):
        self.config = RawConfigParser()

    @mock.patch("tito.tagger.main.strftime", strftime_mock)
    def test_today(self):
        tagger = VersionTagger(self.config)
        assert tagger.today == "Wed Apr 22 2020"

    @unittest.skipIf(PY2, "It is not trivial to represent timezone in python2 because"
                          "datetime.timezone is python3 only and pytz is not in EPEL. This feature"
                          "is for F27+ and EPEL8+ so we don't need to test it for python2")
    @mock.patch("tito.tagger.main.strftime", strftime_mock)
    def test_today_datetime(self):
        self.config["buildconfig"] = {"changelog_date_with_time": True}
        tagger = VersionTagger(self.config)
        assert tagger.today == "Wed Apr 22 15:02:00 UTC 2020"



#
# Copyright (c) 2008-2009 Red Hat, Inc.
#
# This software is licensed to you under the GNU General Public License,
# version 2 (GPLv2). There is NO WARRANTY for this software, express or
# implied, including the implied warranties of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. You should have received a copy of GPLv2
# along with this software; if not, see
# http://www.gnu.org/licenses/old-licenses/gpl-2.0.txt.
#
# Red Hat trademarks are not licensed under GPLv2. No permission is
# granted to use or replicate Red Hat trademarks that are incorporated
# in this software or its documentation.

""" Pure unit tests for tito's common module. """
from tito.common import (replace_version, find_spec_like_file, increase_version,
    search_for, compare_version, run_command_print, find_wrote_in_rpmbuild_output,
    render_cheetah, increase_zstream, reset_release, find_file_with_extension,
    normalize_class_name, extract_sha1, DEFAULT_BUILD_DIR, munge_specfile,
    munge_setup_macro, get_project_name,
    _out)

from tito.compat import StringIO
from tito.tagger import CargoBump

import os
import re
import unittest

from mock import Mock, patch, call
from tempfile import NamedTemporaryFile
from textwrap import dedent
from unit import open_mock, Capture
from blessed import Terminal


class CommonTests(unittest.TestCase):
    def setUp(self):
        # Start in a known location to prevent problems with tests that
        # end in a temp directory that is subsequently deleted.
        os.chdir(DEFAULT_BUILD_DIR)

    def test_normalize_class_name(self):
        """ Test old spacewalk.releng namespace is converted to tito. """
        self.assertEquals("tito.builder.Builder",
                normalize_class_name("tito.builder.Builder"))
        self.assertEquals("tito.builder.Builder",
                normalize_class_name("spacewalk.releng.builder.Builder"))
        self.assertEquals("tito.tagger.VersionTagger",
                normalize_class_name("spacewalk.releng.tagger.VersionTagger"))

    def test_replace_version_leading_whitespace(self):
        line = "    version='1.0'\n"
        expected = "    version='2.5.3'\n"
        self.assertEquals(expected, replace_version(line, "2.5.3"))

    def test_replace_version_no_whitespace(self):
        line = "version='1.0'\n"
        expected = "version='2.5.3'\n"
        self.assertEquals(expected, replace_version(line, "2.5.3"))

    def test_replace_version_some_whitespace(self):
        line = "version = '1.0'\n"
        expected = "version = '2.5.3'\n"
        self.assertEquals(expected, replace_version(line, "2.5.3"))

    def test_replace_version_double_quote(self):
        line = 'version="1.0"\n'
        expected = 'version="2.5.3"\n'
        self.assertEquals(expected, replace_version(line, "2.5.3"))

    def test_replace_version_trailing_chars(self):
        line = "version = '1.0', blah blah blah\n"
        expected = "version = '2.5.3', blah blah blah\n"
        self.assertEquals(expected, replace_version(line, "2.5.3"))

    def test_replace_version_crazy_old_version(self):
        line = "version='1.0asjhd82371kjsdha98475h87asd7---asdai.**&'\n"
        expected = "version='2.5.3'\n"
        self.assertEquals(expected, replace_version(line, "2.5.3"))

    def test_replace_version_crazy_new_version(self):
        line = "version='1.0'\n"
        expected = "version='91asj.;]][[a]sd[]'\n"
        self.assertEquals(expected, replace_version(line,
            "91asj.;]][[a]sd[]"))

    def test_replace_version_uppercase(self):
        line = "VERSION='1.0'\n"
        expected = "VERSION='2.5.3'\n"
        self.assertEquals(expected, replace_version(line, "2.5.3"))

    def test_replace_version_no_match(self):
        line = "this isn't a version fool.\n"
        self.assertEquals(line, replace_version(line, "2.5.3"))

    def test_extract_sha1(self):
        ls_remote_output = "Could not chdir to home directory\n" + \
                           "fe87e2b75ed1850718d99c797cc171b88bfad5ca ref/origin/sometag"
        self.assertEquals("fe87e2b75ed1850718d99c797cc171b88bfad5ca",
                          extract_sha1(ls_remote_output))

    def test_compare_version(self):
        self.assertEquals(0, compare_version("1", "1"))
        self.assertTrue(compare_version("2.1", "2.2") < 0)
        self.assertTrue(compare_version("3.0.4.10", "3.0.4.2") > 0)
        self.assertTrue(compare_version("4.08", "4.08.01") < 0)
        self.assertTrue(compare_version("3.2.1.9.8144", "3.2") > 0)
        self.assertTrue(compare_version("3.2", "3.2.1.9.8144") < 0)
        self.assertTrue(compare_version("1.2", "2.1") < 0)
        self.assertTrue(compare_version("2.1", "1.2") > 0)
        self.assertTrue(compare_version("1.0", "1.0.1") < 0)
        self.assertTrue(compare_version("1.0.1", "1.0") > 0)
        self.assertEquals(0, compare_version("5.6.7", "5.6.7"))
        self.assertEquals(0, compare_version("1.01.1", "1.1.1"))
        self.assertEquals(0, compare_version("1.1.1", "1.01.1"))
        self.assertEquals(0, compare_version("1", "1.0"))
        self.assertEquals(0, compare_version("1.0", "1"))
        self.assertEquals(0, compare_version("1.0.2.0", "1.0.2"))

    def test_run_command_print(self):
        self.assertEquals('', run_command_print("sleep 0.1"))

    def test_rpmbuild_claims_to_be_successful(self):
        succeeded_result = "success"
        output = "Wrote: %s" % succeeded_result

        success_line = find_wrote_in_rpmbuild_output(output)

        self.assertEquals(succeeded_result, success_line[0])

    @patch("tito.common.error_out")
    def test_rpmbuild_which_ended_with_error_is_described_with_the_analyzed_line(self, mock_error):
        output = "some error output from rpmbuild\n" \
            "next error line"

        find_wrote_in_rpmbuild_output(output)

        mock_error.assert_called_once_with("Unable to locate 'Wrote: ' lines in rpmbuild output: '%s'" % output)

    @patch("tito.common.find_file_with_extension")
    def test_find_spec_like_file_tmpl(self, mock_find):
        mock_find.side_effect = [None, "result.spec.tmpl"]
        result = find_spec_like_file()
        self.assertEquals("result.spec.tmpl", result)
        self.assertEquals(2, len(mock_find.mock_calls))

    @patch("tito.common.find_file_with_extension")
    def test_find_spec_like_file_spec(self, mock_find):
        mock_find.side_effect = ["result.spec"]
        result = find_spec_like_file()
        self.assertEquals("result.spec", result)
        self.assertEquals(1, len(mock_find.mock_calls))

    @patch("tito.common.find_file_with_extension")
    def test_find_spec_like_file_no_match(self, mock_find):
        mock_find.side_effect = [None, None]
        with Capture(silent=True):
            self.assertRaises(SystemExit, find_spec_like_file)
            self.assertEquals(2, len(mock_find.mock_calls))

    @patch("os.listdir")
    def test_find_file_with_extension(self, mock_listdir):
        mock_listdir.return_value = ["hello.txt"]
        result = find_file_with_extension("/tmp", ".txt")
        self.assertEquals(mock_listdir.mock_calls[0], call("/tmp"))
        self.assertEquals("/tmp/hello.txt", result)

    @patch("os.listdir")
    def test_find_file_with_extension_no_match(self, mock_listdir):
        mock_listdir.return_value = ["hello.txt"]
        result = find_file_with_extension("/tmp", ".foo")
        self.assertEquals(mock_listdir.mock_calls[0], call("/tmp"))
        self.assertEqual(None, result)

    @patch("os.listdir")
    def test_find_file_with_extension_duplicates(self, mock_listdir):
        mock_listdir.return_value = ["hello.txt", "goodbye.txt"]
        with Capture(silent=True):
            self.assertRaises(SystemExit, find_file_with_extension, "/tmp", ".txt")

    def test_search_for(self):
        content = dedent("""
        HelloWorld
        Hello World
        """)
        with open_mock(content):
            results = search_for("foo", r"(Hello\s+World)", r"(HelloWorld)")
            self.assertEquals(("Hello World",), results[0])
            self.assertEquals(("HelloWorld",), results[1])

    def test_search_for_gets_first_match(self):
        content = dedent("""
        HelloWorld
        Hello World
        """)
        with open_mock(content):
            results = search_for("foo", r"(Hello.*)")
            self.assertEquals(("HelloWorld",), results[0])

    def test_search_for_no_match(self):
        content = dedent("""
        HelloWorld
        Goodbye World
        """)
        with open_mock(content):
            with Capture(silent=True):
                self.assertRaises(SystemExit, search_for, "foo", r"(NoMatch)")

    @patch("tito.common.read_user_config")
    def test_turn_off_colors(self, mock_user_conf):
        mock_user_conf.return_value = {'COLOR': '0'}
        stream = StringIO()
        _out('Hello world', None, Terminal().red, stream)
        self.assertEquals('Hello world\n', stream.getvalue())

    @patch("tito.common.read_user_config")
    def test_colors(self, mock_user_conf):
        mock_user_conf.return_value = {}
        stream = StringIO()
        _out('Hello world', None, Terminal().red, stream)
        # RHEL 6 doesn't have self.assertRegexpMatches unfortunately
        self.assertTrue(re.match('.+Hello world.+\n', stream.getvalue()))

    def test_get_project_name(self):
        TAGS = [
            ('package-1.0-1', 'package'),
            ('package-1.0', 'package'),
            ('long-package-name-that-should-not-be-an-issue-0.1-1', 'long-package-name-that-should-not-be-an-issue'),
            ('package-with-weird-version-0.1-0.1.beta1', 'package-with-weird-version'),
            ('grub2-efi-ia32-1.0-1', 'grub2-efi-ia32'),
            ('iwl5150-firmware-1.0-1', 'iwl5150-firmware'),
            ('389-ds-base-1.0-1', '389-ds-base'),
            ('avr-gcc-c++-1.0-1', 'avr-gcc-c++'),
            ('java-1.8.0-openjdk-1.8.0.232.b09-0', 'java-1.8.0-openjdk'),
            ('jsr-305-0-0.25.20130910svn', 'jsr-305')
        ]

        for (tag, package) in TAGS:
            self.assertEquals(package, get_project_name(tag, None))


class CheetahRenderTest(unittest.TestCase):
    @patch("os.unlink")
    @patch("glob.glob")
    @patch("shutil.move")
    @patch("tito.common.run_command")
    @patch("tempfile.NamedTemporaryFile")
    def test_renders_cheetah(self, mock_tempfile, mock_run_command, mock_move, mock_glob, mock_unlink):
        mock_run_command.return_value = True
        mock_tempfile.return_value.name = "temp_pickle"
        mock_unlink.return_value = True
        mock_glob.return_value = ["/tmp/foo.spec.cheetah"]
        mock_move.return_value = True

        render_cheetah("foo.spec.tmpl", "/tmp", {})
        expected = "cheetah fill --flat --pickle=temp_pickle --odir=/tmp --oext=cheetah foo.spec.tmpl"
        self.assertEquals(call(expected), mock_run_command.mock_calls[0])
        self.assertEquals(call("/tmp/*.cheetah"), mock_glob.mock_calls[0])
        self.assertEquals(call("/tmp/foo.spec.cheetah", "/tmp/foo.spec"), mock_move.mock_calls[0])
        self.assertEquals(call("temp_pickle"), mock_unlink.mock_calls[0])

    @patch("os.unlink")
    @patch("glob.glob")
    @patch("tito.common.run_command")
    @patch("tempfile.NamedTemporaryFile")
    def test_renders_cheetah_missing_result(self, mock_tempfile, mock_run_command, mock_glob, mock_unlink):
        mock_run_command.return_value = True
        mock_tempfile.return_value.name = "temp_pickle"
        mock_unlink.return_value = True
        mock_glob.return_value = []

        with Capture(silent=True):
            self.assertRaises(SystemExit, render_cheetah, "foo.spec.tmpl", "/tmp", {})
            expected = "cheetah fill --flat --pickle=temp_pickle --odir=/tmp --oext=cheetah foo.spec.tmpl"
            self.assertEquals(call(expected), mock_run_command.mock_calls[0])

            self.assertEquals(call("/tmp/*.cheetah"), mock_glob.mock_calls[0])
            self.assertEquals(call("temp_pickle"), mock_unlink.mock_calls[0])


class CargoTransformTest(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_simple_case(self):
        input = ['[package]',
                 'name = "hello_world" # the name of the package',
                 'version = "0.1.0"    # the current version, obeying semver',
                 'authors = ["you@example.com"]']
        output = CargoBump.process_cargo_toml(input, "2.2.2")

        self.assertEquals(4, len(output))
        self.assertEquals("[package]", output[0])
        self.assertEquals("name = \"hello_world\" # the name of the package", output[1])
        self.assertEquals("version = \"2.2.2\"    # the current version, obeying semver", output[2])
        self.assertEquals("authors = [\"you@example.com\"]", output[3])

    def test_complicated_case(self):
        input = ['[package]',
                 'name = "hello_world"',
                 'version = "2.2.2"',
                 'authors = ["you@example.com"]',
                 '',
                 '[dependencies]',
                 'regex = "1.0.0"',
                 '',
                 '[dependencies.termion]',
                 'version = "0.1.0"']
        output = CargoBump.process_cargo_toml(input, "3.3.3")

        self.assertEquals(10, len(output))
        self.assertEquals("version = \"3.3.3\"", output[2])
        self.assertEquals("regex = \"1.0.0\"", output[6])
        self.assertEquals("version = \"0.1.0\"", output[9])


class SpecTransformTest(unittest.TestCase):
    def setUp(self):
        self.spec_file = NamedTemporaryFile(delete=False).name

    def tearDown(self):
        os.unlink(self.spec_file)

    def test_simple_transform(self):
        simple_spec = dedent("""
        Name: Hello
        Version: 1.0.0
        Release: 1%{?dist}
        Source: hello-1.0.0.tar.gz

        %prep
        %setup -q
        """)
        with open(self.spec_file, 'w') as f:
            f.write(simple_spec)

        sha = "acecafe"
        commit_count = 5
        display_version = "git-%s.%s" % (commit_count, sha)
        fullname = "hello-%s" % display_version
        munge_specfile(self.spec_file, sha, commit_count, fullname, "%s.tar.gz" % fullname)
        output = open(self.spec_file, 'r').readlines()

        self.assertEquals(8, len(output))
        self.assertEquals("Release: 1.git.%s.%s%%{?dist}\n" % (commit_count, sha), output[3])
        self.assertEquals("Source: %s.tar.gz\n" % fullname, output[4])
        self.assertEquals("%%setup -q -n %s\n" % fullname, output[7])

        # Spot check some things that should not change
        self.assertEquals("Name: Hello\n", output[1])
        self.assertEqual("%prep\n", output[6])

    def test_transform_release_only(self):
        simple_spec = dedent("""
        Release: 1%{?dist}
        Source: hello-1.0.0.tar.gz
        %setup -q
        """)
        with open(self.spec_file, 'w') as f:
            f.write(simple_spec)

        sha = "acecafe"
        commit_count = 5
        munge_specfile(self.spec_file, sha, commit_count)
        output = open(self.spec_file, 'r').readlines()

        self.assertEquals(4, len(output))
        self.assertEquals("Release: 1.git.%s.%s%%{?dist}\n" % (commit_count, sha), output[1])
        self.assertEquals("Source: hello-1.0.0.tar.gz\n", output[2])
        self.assertEquals("%setup -q\n", output[3])

    def test_transform_no_whitespace_modifications(self):
        simple_spec = dedent("""
        Release:    1%{?dist}
        Source:     hello-1.0.0.tar.gz
        """)
        with open(self.spec_file, 'w') as f:
            f.write(simple_spec)

        sha = "acecafe"
        commit_count = 5
        munge_specfile(self.spec_file, sha, commit_count)
        output = open(self.spec_file, 'r').readlines()

        self.assertEquals(3, len(output))
        self.assertEquals("Release:    1.git.%s.%s%%{?dist}\n" % (commit_count, sha), output[1])
        self.assertEquals("Source:     hello-1.0.0.tar.gz\n", output[2])

    def test_complex_setup_transform(self):
        simple_spec = dedent("""
        %setup -q -n hello-1
        """)
        with open(self.spec_file, 'w') as f:
            f.write(simple_spec)

        sha = "acecafe"
        commit_count = 5
        display_version = "git-%s.%s" % (commit_count, sha)
        fullname = "hello-%s" % display_version
        munge_specfile(self.spec_file, sha, commit_count, fullname, "%s.tar.gz" % fullname)
        output = open(self.spec_file, 'r').readlines()

        self.assertEquals("%%setup -q -n %s\n" % fullname, output[1])

    def test_transform_no_dist_tag(self):
        simple_spec = dedent("""
        Release: 1
        Source: hello-1.0.0.tar.gz
        """)
        with open(self.spec_file, 'w') as f:
            f.write(simple_spec)

        sha = "acecafe"
        commit_count = 5
        munge_specfile(self.spec_file, sha, commit_count)
        output = open(self.spec_file, 'r').readlines()

        self.assertEquals(3, len(output))
        self.assertEquals("Release: 1.git.%s.%s\n" % (commit_count, sha), output[1])
        self.assertEquals("Source: hello-1.0.0.tar.gz\n", output[2])


class VersionMathTest(unittest.TestCase):
    def test_increase_version_minor(self):
        line = "1.0.0"
        expected = "1.0.1"
        self.assertEquals(expected, increase_version(line))

    def test_increase_version_major(self):
        line = "1.0"
        expected = "1.1"
        self.assertEquals(expected, increase_version(line))

    def test_increase_release(self):
        line = "1"
        expected = "2"
        self.assertEquals(expected, increase_version(line))

    def test_underscore_release(self):
        line = "1_PG5"
        expected = "2_PG5"
        self.assertEquals(expected, increase_version(line))

    def test_increase_versionless(self):
        line = "%{app_version}"
        expected = "%{app_version}"
        self.assertEquals(expected, increase_version(line))

    def test_increase_release_with_rpm_cruft(self):
        line = "1%{?dist}"
        expected = "2%{?dist}"
        self.assertEquals(expected, increase_version(line))

    def test_increase_release_with_zstream(self):
        line = "1%{?dist}.1"
        expected = "1%{?dist}.2"
        self.assertEquals(expected, increase_version(line))

    def test_unknown_version(self):
        line = "somethingstrange"
        expected = "somethingstrange"
        self.assertEquals(expected, increase_version(line))

    def test_empty_string(self):
        line = ""
        expected = ""
        self.assertEquals(expected, increase_version(line))

    def test_increase_zstream(self):
        line = "1%{?dist}"
        expected = "1%{?dist}.1"
        self.assertEquals(expected, increase_zstream(line))

    def test_increase_zstream_already_appended(self):
        line = "1%{?dist}.1"
        expected = "1%{?dist}.2"
        self.assertEquals(expected, increase_zstream(line))

    def test_reset_release_with_rpm_cruft(self):
        line = "2%{?dist}"
        expected = "1%{?dist}"
        self.assertEquals(expected, reset_release(line))

    def test_reset_release_with_more_rpm_cruft(self):
        line = "2.beta"
        expected = "1.beta"
        self.assertEquals(expected, reset_release(line))

    def test_reset_release(self):
        line = "2"
        expected = "1"
        self.assertEquals(expected, reset_release(line))


class MungeSetupMacroTests(unittest.TestCase):
    SOURCE = "tito-git-3.20362dd"

    def test_setup(self):
        line = "%setup -q -n tito-%{version}"
        self.assertEqual("%setup -q -n " + self.SOURCE,
                         munge_setup_macro(self.SOURCE, line))

    def test_autosetup(self):
        line = "%autosetup -n tito-%{version}"
        self.assertEqual("%autosetup -n " + self.SOURCE + " -p1",
                         munge_setup_macro(self.SOURCE, line))



# Copyright (c) 2008-2011 Red Hat, Inc.
#
# This software is licensed to you under the GNU General Public License,
# version 2 (GPLv2). There is NO WARRANTY for this software, express or
# implied, including the implied warranties of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. You should have received a copy of GPLv2
# along with this software; if not, see
# http://www.gnu.org/licenses/old-licenses/gpl-2.0.txt.
#
# Red Hat trademarks are not licensed under GPLv2. No permission is
# granted to use or replicate Red Hat trademarks that are incorporated
# in this software or its documentation.
"""
Tito builders for a variety of common methods of building sources, srpms,
and rpms.
"""

import os

from tito.builder import Builder
from tito.common import chdir, debug, run_command, create_tgz, find_spec_like_file, get_commit_timestamp
from tito.tar import TarFixer


class SubmoduleAwareBuilder(Builder):
    """
    Submodule aware builder class.

    This builder is aware of submodules and will package the contents of the submodules as well.
    It expectes `git submodule--helper list` to exist and work.
    """
    REQUIRED_ARGS = []

    def _setup_sources(self):
        """
        Create a copy of the git source for the project at the point in time
        our build tag was created.

        Created in the temporary rpmbuild SOURCES directory.

        *This is the only thing that SubmoduleAwareBuilder changes*.
        *It needs to call a different create_tgz
        """
        self._create_build_dirs()

        debug("Creating %s from git tag: %s..." % (self.tgz_filename,
                                                   self.git_commit_id))

        # call our create_tgz, instead of the global one.
        self.create_tgz(self.git_root, self.tgz_dir, self.git_commit_id,
                        self.relative_project_dir,
                        os.path.join(self.rpmbuild_sourcedir, self.tgz_filename))

        # Extract the source so we can get at the spec file, etc.
        debug("Copying git source to: %s" % self.rpmbuild_gitcopy)
        run_command("cd %s/ && tar xzf %s" % (self.rpmbuild_sourcedir,
                                              self.tgz_filename))

        debug("Show contents of the directory structure we just extracted:\n%s"
              % os.listdir(self.rpmbuild_gitcopy))

        # NOTE: The spec file we actually use is the one exported by git
        # archive into the temp build directory. This is done so we can
        # modify the version/release on the fly when building test rpms
        # that use a git SHA1 for their version.
        self.spec_file_name = os.path.basename(find_spec_like_file(self.rpmbuild_gitcopy))
        self.spec_file = os.path.join(
            self.rpmbuild_gitcopy, self.spec_file_name)

    def run_git_archive(self, relative_git_dir, prefix, commit, dest_tar, subdir):
        # command to generate a git-archive
        git_archive_cmd = 'git archive --format=tar --prefix=%s/ %s:%s --output=%s' % (
            prefix, commit, relative_git_dir, dest_tar)

        with chdir(subdir) as p:
            run_command(git_archive_cmd)

            # Run git-archive separately if --debug was specified.
            # This allows us to detect failure early.
            # On git < 1.7.4-rc0, `git archive ... commit:./` fails!
            debug('git-archive fails if relative dir is not in git tree',
                  '%s > /dev/null' % git_archive_cmd)

    def create_tgz(self, git_root, prefix, commit, relative_dir,
                   dest_tgz):
        """
        Create a .tar.gz from a projects source in git.
        And include submodules
        """

        git_root_abspath = os.path.abspath(git_root)
        gitmodules_path = os.path.join(git_root_abspath, '.gitmodules')

        # if .gitmodules does not exist, just call the existing create_tgz function
        # as there is nothing to see here.
        if not os.path.exists(gitmodules_path):
            return create_tgz(git_root, prefix, commit, relative_dir, dest_tgz)

        os.chdir(git_root_abspath)
        timestamp = get_commit_timestamp(commit)

        # Accommodate standalone projects with specfile in root of git repo:
        relative_git_dir = "%s" % relative_dir
        if relative_git_dir in ['/', './']:
            relative_git_dir = ""

        basename = os.path.splitext(dest_tgz)[0]
        initial_tar = "%s.initial" % basename

        # We need to tar up the following:
        # 1. the current repo
        self.run_git_archive(relative_git_dir, prefix, commit, initial_tar, None)

        # 2. all of the submodules
        # then combine those into a single archive.
        submodules_cmd = 'git submodule--helper list'
        submodules_output = run_command(submodules_cmd)

        # split submodules output on newline
        # then on tab, and the directory is the last entry
        submodules_list = [line.split('\t')[-1] for line in submodules_output.split('\n')]

        submodule_tar_files = [initial_tar]
        # We ignore the hash in the sub modules list as we'll have to get the correct one
        # from the commit id in commit
        for submodule in submodules_list:
            # to find the submodule shars:
            # git rev-parse <commit>:./<submodule>
            rev_parse_cmd = 'git rev-parse %s:./%s' % (commit, submodule)
            submodule_commit = run_command(rev_parse_cmd)
            submodule_tar_file = '%s.%s' % (initial_tar, submodule)
            # prefix should be <prefix>/<submodule>
            submodule_prefix = '%s/%s' % (prefix, submodule)

            self.run_git_archive(relative_git_dir, submodule_prefix, submodule_commit,
                                 submodule_tar_file, submodule)
            submodule_tar_files.append(submodule_tar_file)

        # we need to append all of the submodule tar files onto the initial
        tarfiles = ' '.join(submodule_tar_files)
        run_command("tar -Af %s" % tarfiles)

        fixed_tar = "%s.tar" % basename
        fixed_tar_fh = open(fixed_tar, 'wb')
        try:
            tarfixer = TarFixer(open(initial_tar, 'rb'), fixed_tar_fh, timestamp, commit)
            tarfixer.fix()
        finally:
            fixed_tar_fh.close()

        # It's a pity we can't use Python's gzip, but it doesn't offer an equivalent of -n
        return run_command("gzip -n -c < %s > %s" % (fixed_tar, dest_tgz))



#
# Copyright (c) 2008-2014 Red Hat, Inc.
#
# This software is licensed to you under the GNU General Public License,
# version 2 (GPLv2). There is NO WARRANTY for this software, express or
# implied, including the implied warranties of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. You should have received a copy of GPLv2
# along with this software; if not, see
# http://www.gnu.org/licenses/old-licenses/gpl-2.0.txt.
#
# Red Hat trademarks are not licensed under GPLv2. No permission is
# granted to use or replicate Red Hat trademarks that are incorporated
# in this software or its documentation.
'''
Functional tests to build tito with tito.
This can catch indirect omissions within tito itself.
'''

import os
import shutil
import tempfile
import unittest

from functional.fixture import tito
from glob import glob
from os.path import join


class BuildTitoTests(unittest.TestCase):

    @classmethod
    def setUpClass(self):
        'Run tito build before _all_ tests in this class.'
        self.output_dir = tempfile.mkdtemp("-titotestoutput")
        os.chdir(os.path.abspath(join(__file__, '..', '..', '..')))
        self.artifacts = tito(
            'build --rpm --test --output=%s --offline --no-cleanup --debug' %
            self.output_dir
        )

    @classmethod
    def tearDownClass(self):
        'Clean up after _all_ tests in this class unless any test fails.'
        shutil.rmtree(self.output_dir)

    def test_build_tito(self):
        'Tito creates three artifacts'
        self.assertEqual(3, len(self.artifacts))

    def test_find_srpm(self):
        'One artifact is an SRPM'
        srpms = glob(join(self.output_dir, 'tito-*src.rpm'))
        self.assertEqual(1, len(srpms))

    def test_find_rpm(self):
        'One artifact is a noarch RPM'
        rpms = glob(join(self.output_dir, 'noarch', 'tito-*noarch.rpm'))
        self.assertEqual(1, len(rpms))

    def test_find_tgz(self):
        'One artifact is a tarball'
        tgzs = glob(join(self.output_dir, 'tito-*tar.gz'))
        self.assertEqual(1, len(tgzs))



#
# Copyright (c) 2008-2014 Red Hat, Inc.
#
# This software is licensed to you under the GNU General Public License,
# version 2 (GPLv2). There is NO WARRANTY for this software, express or
# implied, including the implied warranties of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. You should have received a copy of GPLv2
# along with this software; if not, see
# http://www.gnu.org/licenses/old-licenses/gpl-2.0.txt.
#
# Red Hat trademarks are not licensed under GPLv2. No permission is
# granted to use or replicate Red Hat trademarks that are incorporated
# in this software or its documentation.
"""
Functional Tests for the CoprReleaser.
"""

from functional.fixture import TitoGitTestFixture

import mock
from tito.compat import *  # NOQA
from tito.release import CoprReleaser
from unit import Capture

PKG_NAME = "releaseme"

RELEASER_CONF = """
[test]
releaser = tito.release.CoprReleaser
builder = tito.builder.Builder
"""


class CoprReleaserTests(TitoGitTestFixture):

    def setUp(self):
        TitoGitTestFixture.setUp(self)
        self.create_project(PKG_NAME)

        # Setup test config:
        self.config = RawConfigParser()
        self.config.add_section("buildconfig")
        self.config.set("buildconfig", "builder",
                "tito.builder.Builder")
        self.config.set("buildconfig", "offline",
                "true")

        self.releaser_config = RawConfigParser()
        self.releaser_config.add_section('test')
        self.releaser_config.set('test', 'releaser',
            'tito.release.CoprReleaser')
        self.releaser_config.set('test', 'builder',
            'tito.builder.Builder')
        self.releaser_config.set('test', 'project_name', PKG_NAME)
        self.releaser_config.set('test', 'upload_command',
            'scp %(srpm)s example.com/public_html/my_srpm/')
        self.releaser_config.set('test', 'remote_location',
            'http://example.com/~someuser/my_srpm/')

    def test_with_releaser(self):
        releaser = CoprReleaser(PKG_NAME, None, '/tmp/tito/',
            self.config, {}, 'test', self.releaser_config, False,
            False, False, **{'offline': True})
        releaser.release(dry_run=True)
        self.assertTrue(releaser.srpm_submitted is not None)

    def test_with_remote_defined_in_user_conf(self):
        self.releaser_config.remove_option("test", "remote_location")
        user_config = {'COPR_REMOTE_LOCATION': 'http://example.com/~otheruser/'}
        releaser = CoprReleaser(PKG_NAME, None, '/tmp/tito/',
            self.config, user_config, 'test', self.releaser_config, False,
            False, False, **{'offline': True})
        releaser.release(dry_run=True)
        self.assertTrue(releaser.srpm_submitted is not None)

    @mock.patch("tito.release.CoprReleaser._submit")
    @mock.patch("tito.release.CoprReleaser._upload")
    def test_no_remote_defined(self, upload, submit):
        self.releaser_config.remove_option("test", "remote_location")
        releaser = CoprReleaser(PKG_NAME, None, '/tmp/tito/',
            self.config, {}, 'test', self.releaser_config, False,
            False, False, **{'offline': True})
        releaser.release(dry_run=True)

        self.assertFalse(upload.called)
        self.assertTrue(submit.called)

    @mock.patch("tito.release.CoprReleaser._run_command")
    def test_multiple_project_names(self, run_command):
        self.releaser_config.remove_option("test", "remote_location")
        self.releaser_config.set('test', 'project_name', "%s %s" % (PKG_NAME,
            PKG_NAME))
        releaser = CoprReleaser(PKG_NAME, None, '/tmp/tito/',
            self.config, {}, 'test', self.releaser_config, False,
            False, False, **{'offline': True})
        releaser.release()
        args = mock.call('/usr/bin/copr-cli build --nowait releaseme %s' %
                releaser.builder.srpm_location)
        self.assertEquals(args, run_command.mock_calls[0])
        self.assertEquals(args, run_command.mock_calls[1])



#
# Copyright (c) 2008-2010 Red Hat, Inc.
#
# This software is licensed to you under the GNU General Public License,
# version 2 (GPLv2). There is NO WARRANTY for this software, express or
# implied, including the implied warranties of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. You should have received a copy of GPLv2
# along with this software; if not, see
# http://www.gnu.org/licenses/old-licenses/gpl-2.0.txt.
#
# Red Hat trademarks are not licensed under GPLv2. No permission is
# granted to use or replicate Red Hat trademarks that are incorporated
# in this software or its documentation.

import os
import sys
import shutil
import tempfile
import unittest

from tito.cli import CLI
from tito.common import run_command

# NOTE: No Name in test spec file as we re-use it for several packages.
# Name must be written first.
TEST_SPEC_3 = """
%{!?python3_sitelib: %define python3_sitelib %(%{__python3} -c "from distutils.sysconfig import get_python_lib; print get_python_lib()")}

Version:        0.0.1
Release:        1%{?dist}
Summary:        Tito test package.
URL:            https://example.com
Group:          Applications/Internet
License:        GPLv2
BuildRoot:      %{_tmppath}/%{name}-root-%(%{__id_u} -n)
BuildArch:      noarch
BuildRequires:  python3-devel
BuildRequires:  python3-setuptools
Source0:        %{name}-%{version}.tar.gz

%description
Nobody cares.

%prep
#nothing to do here
%setup -q -n %{name}-%{version}

%build
%{__python3} setup.py build

%install
rm -rf $RPM_BUILD_ROOT
%{__python3} setup.py install -O1 --skip-build --root $RPM_BUILD_ROOT
rm -f $RPM_BUILD_ROOT%{python3_sitelib}/*egg-info/requires.txt

%clean
rm -rf %{buildroot}

%files
%defattr(-,root,root)
#%dir %{python3_sitelib}/%{name}
%{python3_sitelib}/%{name}-*.egg-info

%changelog
"""

# NOTE: No Name in test spec file as we re-use it for several packages.
# Name must be written first.
TEST_SPEC_2 = """
%{!?python_sitelib: %define python_sitelib %(%{__python} -c "from distutils.sysconfig import get_python_lib; print get_python_lib()")}

Version:        0.0.1
Release:        1%{?dist}
Summary:        Tito test package.
URL:            https://example.com
Group:          Applications/Internet
License:        GPLv2
BuildRoot:      %{_tmppath}/%{name}-root-%(%{__id_u} -n)
BuildArch:      noarch
BuildRequires:  python-devel
BuildRequires:  python-setuptools
Source0:        %{name}-%{version}.tar.gz

%description
Nobody cares.

%prep
#nothing to do here
%setup -q -n %{name}-%{version}

%build
%{__python} setup.py build

%install
rm -rf $RPM_BUILD_ROOT
%{__python} setup.py install -O1 --skip-build --root $RPM_BUILD_ROOT
rm -f $RPM_BUILD_ROOT%{python_sitelib}/*egg-info/requires.txt

%clean
rm -rf %{buildroot}

%files
%defattr(-,root,root)
#%dir %{python_sitelib}/%{name}
%{python_sitelib}/%{name}-*.egg-info

%changelog
"""

TEST_SETUP_PY = """
from setuptools import setup, find_packages

setup(
    name="%s",
    version='1.0',
    description='tito test project',
    author='Nobody Knows',
    author_email='tito@example.com',
    url='http://rm-rf.ca/tito',
    license='GPLv2+',

    package_dir={
        '%s': 'src/',
    },
    packages = find_packages('src'),
    include_package_data = True,

    classifiers = [
        'License :: OSI Approved :: GNU General Public License (GPL)',
        'Development Status :: 2 - Pre-Alpha',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: Information Technology',
        'Programming Language :: Python'
    ],
)
"""

TEST_PYTHON_SRC = """
class Empty(object):
    pass
"""


def tito(argstring):
    """ Run Tito from source with given arguments. """
    return CLI().main(argstring.split(' '))


class TitoGitTestFixture(unittest.TestCase):
    """
    Fixture providing setup/teardown and utilities for all tests requiring
    an actual git repository.
    """
    def setUp(self):
        # Create a temporary directory for our test git repository:
        self.repo_dir = tempfile.mkdtemp("-titotest")
        print
        print
        print("Testing in: %s" % self.repo_dir)
        print

        # Initialize the repo:
        os.chdir(self.repo_dir)
        run_command('git init')

        # Next we tito init:
        tito("init")
        run_command('echo "offline = true" >> .tito/tito.props')
        run_command('git add .tito/tito.props')
        run_command("git commit -m 'set offline in tito.props'")

    def tearDown(self):
        run_command('chmod -R u+rw %s' % self.repo_dir)
        pass

    def write_file(self, path, contents):
        out_f = open(path, 'w')
        out_f.write(contents)
        out_f.close()

    def create_project_from_spec(self, pkg_name, config,
            pkg_dir='', spec=None):
        """
        Create a sample tito project and copy the given test spec file over.
        """
        full_pkg_dir = os.path.join(self.repo_dir, pkg_dir)
        run_command('mkdir -p %s' % full_pkg_dir)
        os.chdir(full_pkg_dir)

        shutil.copyfile(spec, os.path.join(full_pkg_dir, os.path.basename(spec)))

        # Write the config object we were given out to the project repo:
        configfile = open(os.path.join(full_pkg_dir, 'tito.props'), 'w')
        config.write(configfile)
        configfile.close()

    def create_project(self, pkg_name, pkg_dir='', tag=True):
        """
        Create a test project at the given location, assumed to be within
        our test repo, but possibly within a sub-directory.
        """
        full_pkg_dir = os.path.join(self.repo_dir, pkg_dir)
        run_command('mkdir -p %s' % full_pkg_dir)
        os.chdir(full_pkg_dir)

        # TODO: Test project needs work, doesn't work in some scenarios
        # like UpstreamBuilder:
        self.write_file(os.path.join(full_pkg_dir, 'a.txt'), "BLERG\n")

        # Write the test spec file:
        spec = TEST_SPEC_3 if sys.version_info[0] == 3 else TEST_SPEC_2
        self.write_file(os.path.join(full_pkg_dir, "%s.spec" % pkg_name),
            "Name: %s\n%s" % (pkg_name, spec))

        # Write test setup.py:
        self.write_file(os.path.join(full_pkg_dir, "setup.py"),
            TEST_SETUP_PY % (pkg_name, pkg_name))

        # Write test source:
        run_command('mkdir -p %s' % os.path.join(full_pkg_dir, "src"))
        self.write_file(os.path.join(full_pkg_dir, "src", "module.py"),
            TEST_PYTHON_SRC)

        files = [os.path.join(pkg_dir, 'a.txt'),
                os.path.join(pkg_dir, 'setup.py'),
                os.path.join(pkg_dir, '%s.spec' % pkg_name),
                os.path.join(pkg_dir, 'src/module.py')
        ]
        run_command('git add %s' % ' '.join(files))
        run_command("git commit -m 'initial commit'")

        if tag:
            tito('tag --keep-version --debug --accept-auto-changelog')
